"""
Tests the asilib's Conjunction class.
"""
from datetime import datetime, timedelta
import dateutil.parser
import string

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.testing.decorators
import pandas as pd
import pytest
import numpy as np

try:
    import IRBEM

    irbem_imported = True
except ImportError as err:
    # You can also get a ModuleNotFoundError if IRBEM is not installed
    # (as compared to failed to import), but it is a subclass of ImportError.
    irbem_imported = False

import asilib
import asilib.asi.fake_asi
import asilib.asi
from asilib.imager import Skymap_Cleaner
from .mock_footprint import footprint


t0 = dateutil.parser.parse('2014-05-05T04:49:10')
location_code = 'gill'


def test_conjunction_find_none():
    """
    Verifies that no start or end conjunction intervals are identified.
    """
    asi = asilib.asi.themis(location_code, time=t0, load_images=False, alt=110)
    times, lla = footprint(asi.meta['lon'] + 100)
    c = asilib.Conjunction(asi, (times, lla))
    df = c.find()
    assert df.shape == (0, 4)
    return


def test_conjunction_find_multiple():
    """
    Verifies that multiple start and end conjunction intervals are identified.
    """
    asi = asilib.asi.themis(location_code, time=t0, load_images=False, alt=110)
    times, lla = footprint(asi.meta['lon'], alt=110)

    c = asilib.Conjunction(asi, (times, lla))
    df = c.find(min_elevation=20)
    assert df.shape == (18, 4)
    assert np.all(
        df.to_numpy()
        == np.array(
            [
            [
                pd.Timestamp('2015-01-01 01:44:42'),
                pd.Timestamp('2015-01-01 01:45:42'), 1047, 1057
            ],
            [   pd.Timestamp('2015-01-01 02:11:48'),
                pd.Timestamp('2015-01-01 02:12:48'), 1318, 1328
            ],
            [   pd.Timestamp('2015-01-01 03:19:42'),
                pd.Timestamp('2015-01-01 03:20:42'), 1997, 2007
            ],
            [   pd.Timestamp('2015-01-01 03:46:48'),
                pd.Timestamp('2015-01-01 03:47:48'), 2268, 2278
            ],
            [   pd.Timestamp('2015-01-01 04:54:42'),
                pd.Timestamp('2015-01-01 04:55:42'), 2947, 2957
            ],
            [   pd.Timestamp('2015-01-01 05:21:48'),
                pd.Timestamp('2015-01-01 05:22:48'), 3218, 3228
            ],
            [   pd.Timestamp('2015-01-01 06:29:42'),
                pd.Timestamp('2015-01-01 06:30:42'), 3897, 3907
            ],
            [   pd.Timestamp('2015-01-01 06:56:48'),
                pd.Timestamp('2015-01-01 06:57:48'), 4168, 4178
            ],
            [   pd.Timestamp('2015-01-01 08:04:42'),
                pd.Timestamp('2015-01-01 08:05:42'), 4847, 4857
            ],
            [   pd.Timestamp('2015-01-01 08:31:48'),
                pd.Timestamp('2015-01-01 08:32:48'), 5118, 5128
            ],
            [   pd.Timestamp('2015-01-01 09:39:42'),
                pd.Timestamp('2015-01-01 09:40:42'), 5797, 5807
            ],
            [   pd.Timestamp('2015-01-01 10:06:48'),
                pd.Timestamp('2015-01-01 10:07:48'), 6068, 6078
            ],
            [   pd.Timestamp('2015-01-01 11:14:42'),
                pd.Timestamp('2015-01-01 11:15:42'), 6747, 6757
            ],
            [   pd.Timestamp('2015-01-01 11:41:48'),
                pd.Timestamp('2015-01-01 11:42:48'), 7018, 7028
            ],
            [   pd.Timestamp('2015-01-01 12:49:42'),
                pd.Timestamp('2015-01-01 12:50:42'), 7697, 7707
            ],
            [   pd.Timestamp('2015-01-01 13:16:48'),
                pd.Timestamp('2015-01-01 13:17:48'), 7968, 7978
            ],
            [   pd.Timestamp('2015-01-01 14:24:42'),
                pd.Timestamp('2015-01-01 14:25:42'), 8647, 8657
            ],
            [
                pd.Timestamp('2015-01-01 14:51:48'),
                pd.Timestamp('2015-01-01 14:52:48'), 8918, 8928
            ]
            ],
            dtype=object,
        )
    )
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_conjunction_find_multiple'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_plot_conjunction_find_multiple():
    """
    Plots the ASI map and superposes the footprint start and end indices.
    """
    asi = asilib.asi.themis(location_code, time=t0, load_images=False, alt=110)
    times, lla = footprint(asi.meta['lon'], alt=110)

    c = asilib.Conjunction(asi, (times, lla))
    df = c.find(min_elevation=20)

    _cleaner = Skymap_Cleaner(asi.skymap['lon'], asi.skymap['lat'], asi.skymap['el'])
    _cleaner.mask_elevation(20)
    _lon_map, _lat_map = _cleaner.remove_nans()

    _, ax = plt.subplots()    
    ax.pcolormesh(_lon_map, _lat_map, np.ones_like(_lat_map))
    ax.plot(lla[:, 1], lla[:, 0], 'k')
    ax.scatter(lla[df['start_index'], 1], lla[df['start_index'], 0], c='g', s=100)
    ax.scatter(lla[df['end_index'], 1], lla[df['end_index'], 0], c='c', s=100)

    ax.set(
        xlim=(np.nanmin(c._lon_map) - 5, np.nanmax(c._lon_map) + 5),
        ylim=(np.nanmin(c._lat_map) - 2, np.nanmax(c._lat_map) + 2),
    )
    return


def test_interp_sat():
    """
    Test if the satellite LLA is correctly interpolated and aligned to the ASI timestamps.

    The satellite LLA timestamps every 6 seconds, while THEMIS ASI timestamps are every
    3 seconds.
    """
    asi_time_range = (t0, t0 + timedelta(minutes=1))
    # Add 6 seconds so that the footprint interval completely encompasses the
    # asi time_range.
    footprint_time_range = (t0, t0 + timedelta(minutes=1, seconds=6))
    asi = asilib.asi.themis(location_code, time_range=asi_time_range, alt=110)
    times, lla = footprint(asi.meta['lon'], time_range=footprint_time_range)
    c = asilib.Conjunction(asi, (times, lla))

    c.interp_sat()
    assert c.sat.shape == (20, 3)
    assert np.all(c.sat.index == c.imager.data.time)
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_interp_sat'], tol=10, remove_text=True, extensions=['png']
)
def test_plot_interp_sat():
    """
    Test if the satellite LLA is correctly interpolated and aligned to the ASI timestamps.

    The satellite LLA timestamps every 6 seconds, while THEMIS ASI timestamps are every
    3 seconds.
    """
    asi_time_range = (t0, t0 + timedelta(minutes=1))
    # Add 6 seconds so that the footprint interval completely encompasses the
    # asi time_range.
    footprint_time_range = (t0, t0 + timedelta(minutes=1, seconds=6))
    asi = asilib.asi.themis(location_code, time_range=asi_time_range, alt=110)
    times, lla = footprint(asi.meta['lon'], time_range=footprint_time_range)
    c = asilib.Conjunction(asi, (times, lla))

    c.interp_sat()

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].scatter(times, lla[:, 0], c='k', s=50)
    ax[0].scatter(c.sat.index, c.sat['lat'], c='orange')

    ax[1].scatter(times, lla[:, 1], c='k', s=50)
    ax[1].scatter(c.sat.index, c.sat['lon'], c='orange')

    ax[0].axvline(c.imager.data.time[0])
    ax[0].axvline(c.imager.data.time[-1])
    return


@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_plot_interp_sat_wrap'], tol=10, remove_text=True, extensions=['png']
)
def test_plot_interp_sat_wrap():
    """
    Test if the satellite LLA is correctly interpolated and aligned to the ASI timestamps
    when the satellite's longitudes pass through the 180 meridian.

    The satellite LLA timestamps every 6 seconds, while THEMIS ASI timestamps are every
    3 seconds.
    """
    asi_time_range = (t0, t0 + timedelta(minutes=1))
    # Add 6 seconds so that the footprint interval completely encompasses the
    # asi time_range.
    footprint_time_range = (t0, t0 + timedelta(minutes=1, seconds=6))
    asi = asilib.asi.themis(location_code, time_range=asi_time_range, alt=110)
    times, lla = footprint(-180, time_range=footprint_time_range, precession_rate=20)
    lla[lla[:, 1] < -180, 1] += 360
    c = asilib.Conjunction(asi, (times, lla))

    with pytest.warns(UserWarning):
        c.interp_sat()

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].scatter(times, lla[:, 0], c='k', s=50)
    ax[0].scatter(c.sat.index, c.sat['lat'], c='orange')

    ax[1].scatter(times, lla[:, 1], c='k', s=50)
    ax[1].scatter(c.sat.index, c.sat['lon'], c='orange')

    ax[0].axvline(c.imager.data.time[0])
    ax[0].axvline(c.imager.data.time[-1])
    return


@pytest.mark.skipif(not irbem_imported, reason='IRBEM is not installed.')
def test_magnetic_tracing():
    location = 'ATHA'
    dt = 10  # seconds
    time_range = (datetime(2020, 1, 1, 0, 0, 0), datetime(2020, 1, 1, 0, 1, 0))
    asi = asilib.asi.fake_asi.fake_asi(location, time_range=time_range)

    n = int((time_range[1]-time_range[0]).total_seconds()//dt)
    sat_time = np.array([time_range[0] + timedelta(seconds=i*dt) 
                         for i in range(n)])
    lats = np.linspace(asi.meta['lat'] - 2, asi.meta['lat'] + 2, num=sat_time.shape[0])
    lons = np.linspace(asi.meta['lon'] - 2, asi.meta['lon'] + 2, num=sat_time.shape[0])
    sat_lla = np.stack((lats, lons, 500 * np.ones_like(lons))).T
    c = asilib.Conjunction(asi, (sat_time, sat_lla))

    c.lla_footprint(110)
    
    assert np.all(c.sat.index == sat_time)
    np.testing.assert_almost_equal(c.sat['alt'], 110*np.ones_like(c.sat['alt']), decimal=1)
    np.testing.assert_almost_equal(
        c.sat['lat'],
        np.array([53.63169896, 54.39262177, 55.1545005 , 55.91733138, 56.68123125,
            57.44607773])
    )
    np.testing.assert_almost_equal(
        c.sat['lon'],
        np.array([-114.9137138, -114.12793  , -113.342572 , -112.5576476,
            -111.7731121, -110.9890267])
    )
    return


def test_azel_single_lla():
    """
    Tests that one LLA input, right above the imager, is mapped 90 degree elevation
    and near the pixel midpoint (it won't be exactly in the middle).
    """
    location = 'ATHA'
    time = datetime(2020, 1, 1)
    asi = asilib.asi.fake_asi.fake_asi(location, time=time)

    sat_lla = np.array([[asi.meta['lat'], asi.meta['lon'], 500]])
    sat_time = np.array([time])
    c = asilib.Conjunction(asi, (sat_time, sat_lla))
    azel, pixels = c.map_azel()

    # Test the El values
    assert azel[0, 1] == 90

    # Test that the AzEl indices are within a pixel of zenith.
    assert np.max(abs(pixels[0, :] - asi.meta['resolution'][0] / 2)) <= 1
    return


def test_azel_multiple_lla():
    """
    Tests that multiple LLA inputs, nearby and above the imager, are mapped correctly.
    """
    location = 'ATHA'
    time = datetime(2020, 1, 1)
    asi = asilib.asi.fake_asi.fake_asi(location, time=time)

    num = 10
    lats = np.linspace(asi.meta['lat'] - 2, asi.meta['lat'] + 2, num=num)
    lons = np.linspace(asi.meta['lon'] - 2, asi.meta['lon'] + 2, num=num)
    sat_lla = np.stack((lats, lons, 500 * np.ones_like(lons))).T
    times = np.repeat(np.array([time]), num)
    c = asilib.Conjunction(asi, (times, sat_lla))
    azel, pixels = c.map_azel()

    assert np.all(
        np.isclose(
            azel,
            np.array(
                [
                    [211.47839446, 60.51190213],
                    [211.16566566, 66.37949079],
                    [210.85158923, 72.72795178],
                    [210.53615933, 79.4686885],
                    [210.2193701, 86.46258136],
                    [29.90121566, 86.4673325],
                    [29.58169008, 79.51031894],
                    [29.26078743, 72.8378054],
                    [28.93850173, 66.57984562],
                    [28.61482701, 60.81521379],
                ]
            ),
        )
    )

    assert np.all(
        np.isclose(
            pixels,
            np.array(
                [
                    [310.0, 167.0],
                    [303.0, 178.0],
                    [295.0, 190.0],
                    [285.0, 205.0],
                    [273.0, 227.0],
                    [239.0, 285.0],
                    [226.0, 306.0],
                    [218.0, 322.0],
                    [212.0, 335.0],
                    [206.0, 346.0],
                ]
            ),
        )
    )
    return


def test_intensity_closest_pixel():
    """
    Test the auroral intensity from the nearest pixel.
    """
    location_code = 'RANK'
    alt = 110  # km
    time_range = (datetime(2017, 9, 15, 2, 32, 0), datetime(2017, 9, 15, 2, 35, 0))

    asi = asilib.asi.themis(location_code, time_range=time_range, alt=alt)

    # Create the fake satellite track coordinates: latitude, longitude, altitude (LLA).
    # This is a north-south satellite track oriented to the east of the THEMIS/RANK
    # imager.
    n = int((time_range[1] - time_range[0]).total_seconds() / 3)  # 3 second cadence.
    lats = np.linspace(asi.meta["lat"] + 5, asi.meta["lat"] - 5, n)
    lons = (asi.meta["lon"] - 0.5) * np.ones(n)
    alts = alt * np.ones(n)  # Altitude needs to be the same as the skymap.
    sat_lla = np.array([lats, lons, alts]).T
    # Normally the satellite time stamps are not the same as the ASI.
    # You may need to call Conjunction.interp_sat() to find the LLA coordinates
    # at the ASI timestamps.
    sat_time = asi.data.time

    conjunction_obj = asilib.Conjunction(asi, (sat_time, sat_lla))
    nearest_pixel_intensity = conjunction_obj.intensity(box=None)
    assert np.all(
        np.isclose(
            nearest_pixel_intensity,
            np.array(
                [
                    4328.0,
                    4298.0,
                    4258.0,
                    4252.0,
                    4288.0,
                    4211.0,
                    4118.0,
                    4120.0,
                    4089.0,
                    4004.0,
                    4060.0,
                    4005.0,
                    3991.0,
                    3908.0,
                    3890.0,
                    3887.0,
                    3824.0,
                    3757.0,
                    3678.0,
                    3643.0,
                    3644.0,
                    3658.0,
                    3643.0,
                    3548.0,
                    3574.0,
                    3565.0,
                    3526.0,
                    3534.0,
                    3480.0,
                    3494.0,
                    3933.0,
                    12374.0,
                    16140.0,
                    5523.0,
                    7557.0,
                    12812.0,
                    5121.0,
                    4202.0,
                    4051.0,
                    4101.0,
                    4015.0,
                    4198.0,
                    4136.0,
                    4331.0,
                    4306.0,
                    4368.0,
                    4164.0,
                    4462.0,
                    4541.0,
                    4515.0,
                    4612.0,
                    4705.0,
                    4861.0,
                    4960.0,
                    4959.0,
                    5036.0,
                    5098.0,
                    5092.0,
                    5158.0,
                    5295.0,
                ]
            ),
        )
    )
    return

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_rgb_intensity_closest_pixel'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_rgb_intensity_closest_pixel():
    """
    Test the RGB auroral intensity from the nearest pixel.
    """
    time_range = (datetime(2021, 11, 4, 7, 2, 0), datetime(2021, 11, 4, 7, 4, 0))
    location_code = 'LUCK'
    alt = 110

    asi = asilib.asi.trex_rgb(location_code, time_range=time_range, alt=alt)

    # Create a orbit with a footprint every 5 seconds.
    dt_sat = 5
    n_sat_times = int((time_range[1]-time_range[0]).total_seconds()//dt_sat)
    sat_times = np.array([time_range[0]+timedelta(seconds=dt_sat*i) for i in range(n_sat_times)])
    lats = np.linspace(asi.meta['lat']-3, asi.meta['lat']+3, num=n_sat_times)
    lons = (asi.meta['lon']-2) * np.linspace(1.02, 0.98, num=n_sat_times)
    alts = alt * np.ones_like(lats)
    lla = np.stack((lats, lons, alts)).T

    c = asilib.Conjunction(asi, (sat_times, lla))
    intensity = c.intensity()

    plot_n_times = 4
    fig = plt.figure(figsize=(3*plot_n_times, 6))
    gs = gridspec.GridSpec(2, plot_n_times)
    ax = [fig.add_subplot(gs[0, i]) for i in range(plot_n_times)]
    bx = fig.add_subplot(gs[1, :])
    dt_plot = int((time_range[1]-time_range[0]).total_seconds()//plot_n_times)

    for i, ax_i in enumerate(ax):
        t_i = time_range[0] + timedelta(seconds=(i+0.5)*dt_plot)
        asi_image = asi[t_i]
        asi_image.plot_map(ax=ax_i)
        ax_i.text(
            0, 0.99, f'({string.ascii_uppercase[i]}) {asi_image.file_info["time"]}', 
            transform=ax_i.transAxes, va='top', fontsize=12,
            bbox=dict(facecolor='white', edgecolor='red', pad=0.01),
            )
        ax_i.plot(lla[:, 1], lla[:, 0], c='r', ls=':')

        idt_min = np.argmin(np.abs([sat_time_i - t_i for sat_time_i in sat_times]))
        ax_i.scatter(lla[idt_min, 1], lla[idt_min, 0], c='r', marker='x')
        bx.axvline(sat_times[idt_min], c='k')

    for color, _intensity in zip(['r', 'g', 'b'], intensity.T):
        bx.plot(c.sat.index, _intensity, c=color)
    fig.tight_layout()
    return

def test_intensity_area():
    """
    Test the mean auroral intensity in a 10x10 km box.
    """
    # ASI parameters
    location_code = 'RANK'
    alt = 110  # km
    time_range = (datetime(2017, 9, 15, 2, 32, 0), datetime(2017, 9, 15, 2, 35, 0))

    asi = asilib.asi.themis(location_code, time_range=time_range, alt=alt)

    # Create the fake satellite track coordinates: latitude, longitude, altitude (LLA).
    # This is a north-south satellite track oriented to the east of the THEMIS/RANK
    # imager.
    n = int((time_range[1] - time_range[0]).total_seconds() / 3)  # 3 second cadence.
    lats = np.linspace(asi.meta["lat"] + 5, asi.meta["lat"] - 5, n)
    lons = (asi.meta["lon"] - 0.5) * np.ones(n)
    alts = alt * np.ones(n)  # Altitude needs to be the same as the skymap.
    sat_lla = np.array([lats, lons, alts]).T
    # Normally the satellite time stamps are not the same as the ASI.
    # You may need to call Conjunction.interp_sat() to find the LLA coordinates
    # at the ASI timestamps.
    sat_time = asi.data.time

    conjunction_obj = asilib.Conjunction(asi, (sat_time, sat_lla))
    area_intensity = conjunction_obj.intensity(box=(10, 10))

    assert np.all(
        np.isclose(
            area_intensity,
            np.array(
                [
                    4429.5,
                    4307.0,
                    4359.0,
                    4258.0,
                    4277.0,
                    4307.0,
                    4142.66666667,
                    4129.0,
                    4092.0,
                    4002.0,
                    3991.66666667,
                    3990.0,
                    3987.0,
                    3906.66666667,
                    3895.33333333,
                    3909.0,
                    3829.75,
                    3774.0,
                    3741.75,
                    3699.8,
                    3685.88888889,
                    3616.9,
                    3615.3125,
                    3585.3125,
                    3548.33333333,
                    3532.23076923,
                    3501.77142857,
                    3496.58536585,
                    3485.09803922,
                    3532.40740741,
                    3904.0,
                    11476.43137255,
                    15238.31818182,
                    5658.125,
                    7793.53571429,
                    12330.76190476,
                    5629.0,
                    4244.69230769,
                    4105.81818182,
                    4064.2,
                    4055.14285714,
                    4131.33333333,
                    4180.16666667,
                    4236.4,
                    4297.33333333,
                    4354.66666667,
                    4291.0,
                    4386.0,
                    4530.5,
                    4537.0,
                    4605.5,
                    4627.0,
                    4711.0,
                    4799.5,
                    4959.0,
                    5078.0,
                    5014.0,
                    5007.5,
                    5070.0,
                    5156.0,
                ]
            ),
        )
    )
    return

@matplotlib.testing.decorators.image_comparison(
    baseline_images=['test_rgb_intensity_area'],
    tol=10,
    remove_text=True,
    extensions=['png'],
)
def test_rgb_intensity_area():
    """
    Test the RGB auroral intensity in a (10, 10) km area.
    """
    time_range = (datetime(2021, 11, 4, 7, 2, 0), datetime(2021, 11, 4, 7, 4, 0))
    location_code = 'LUCK'
    alt = 110

    asi = asilib.asi.trex_rgb(location_code, time_range=time_range, alt=alt)

    # Create a orbit with a footprint every 5 seconds.
    dt_sat = 5
    n_sat_times = int((time_range[1]-time_range[0]).total_seconds()//dt_sat)
    sat_times = np.array([time_range[0]+timedelta(seconds=dt_sat*i) for i in range(n_sat_times)])
    lats = np.linspace(asi.meta['lat']-3, asi.meta['lat']+3, num=n_sat_times)
    lons = (asi.meta['lon']-2) * np.linspace(1.02, 0.98, num=n_sat_times)
    alts = alt * np.ones_like(lats)
    lla = np.stack((lats, lons, alts)).T

    c = asilib.Conjunction(asi, (sat_times, lla))
    intensity = c.intensity(box=(10,10))

    plot_n_times = 4
    fig = plt.figure(figsize=(3*plot_n_times, 6))
    gs = gridspec.GridSpec(2, plot_n_times)
    ax = [fig.add_subplot(gs[0, i]) for i in range(plot_n_times)]
    bx = fig.add_subplot(gs[1, :])
    dt_plot = int((time_range[1]-time_range[0]).total_seconds()//plot_n_times)

    for i, ax_i in enumerate(ax):
        t_i = time_range[0] + timedelta(seconds=(i+0.5)*dt_plot)
        asi_image = asi[t_i]
        asi_image.plot_map(ax=ax_i)
        ax_i.text(
            0, 0.99, f'({string.ascii_uppercase[i]}) {asi_image.file_info["time"]}', 
            transform=ax_i.transAxes, va='top', fontsize=12,
            bbox=dict(facecolor='white', edgecolor='red', pad=0.01),
            )
        ax_i.plot(lla[:, 1], lla[:, 0], c='r', ls=':')

        idt_min = np.argmin(np.abs([sat_time_i - t_i for sat_time_i in sat_times]))
        ax_i.scatter(lla[idt_min, 1], lla[idt_min, 0], c='r', marker='x')
        bx.axvline(sat_times[idt_min], c='k')

    for color, _intensity in zip(['r', 'g', 'b'], intensity.T):
        bx.plot(c.sat.index, _intensity, c=color)
    fig.tight_layout()
    return


import sys
import pathlib
import configparser

# Run the configuration script when the user runs
# python3 -m asilib [init, initialize, config, or configure]

here = pathlib.Path(__file__).parent.resolve()


if (len(sys.argv) > 1) and (sys.argv[1] in ['init', 'initialize', 'config', 'configure']):
    print('Running the configuration script.')
    s = (
        f'What is the aurora data directory? Press enter for the default '
        f'directory at ~/asilib-data folder will be created.\n'
    )
    ASI_DATA_DIR = input(s).strip()

    # If the user specified the directory, check that the ASI directory already exists
    # and make that directory if it does not.
    if ASI_DATA_DIR != '':
        if not pathlib.Path(ASI_DATA_DIR).exists():
            pathlib.Path(ASI_DATA_DIR).mkdir(parents=True)
            print(f'Made aurora_asi data directory at {pathlib.Path(ASI_DATA_DIR)}.')
        else:
            print(f'aurora_asi data directory at {pathlib.Path(ASI_DATA_DIR)} already exists.')
    else:
        # If the user did not specify the directory, make one at ~/asilib-data
        # and don't save ASI_DATA_DIR. Then configparser in __init__.py will
        # try to load ASI_DATA_DIR and default to pathlib.Path.home() / 'asilib-data'
        # if it doesn't exist.
        DEFAULT_ASI_DATA_DIR = pathlib.Path.home() / 'asilib-data'
        if not DEFAULT_ASI_DATA_DIR.exists():
            DEFAULT_ASI_DATA_DIR.mkdir()
            print(f'asilib directory at {DEFAULT_ASI_DATA_DIR} created.')
        else:
            print(f'asilib directory at {DEFAULT_ASI_DATA_DIR} already exists.')

    # Create a configparser object and add the user configuration.
    config = configparser.ConfigParser()
    if ASI_DATA_DIR != '':
        config['Paths'] = {'ASI_DATA_DIR': ASI_DATA_DIR}

    with open(here / 'config.ini', 'w') as f:
        config.write(f)

else:
    print(
        'This is a configuration script to set up config.ini file. The config '
        'file contains the aurora data directory, and is in ~/asilib-data by '
        'default. To configure this package, run python3 -m asilib config'
    )



import asilib
import asilib.asi

time_range = ('2021-11-04T06:55', '2021-11-04T07:05')
asis = asilib.Imagers(
    [asilib.asi.trex_rgb(location_code, time_range=time_range)
    for location_code in ['LUCK', 'PINA', 'GILL', 'RABB']]
    )
asis.animate_map(lon_bounds=(-115, -85), lat_bounds=(43, 63), overwrite=True)


"""
asilib.Imager provides methods to load ASI time stamps and images, as well as
basic plotting methods.

.. note::
    Most of the basic asilib operations are to download and load large amounts
    of images. Therefore, your performance is largely impacted by your internet speed
    and the type of hard drive/memory on your machine. 
"""
import datetime
import dateutil.parser
import pathlib
import inspect
import shutil
import copy
from collections import namedtuple
from typing import List, Tuple, Generator, Union, Iterable
import warnings
import operator

import numpy as np
import numpy.linalg
import numpy.polynomial
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import ffmpeg
import aacgmv2

try:
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature
    import cartopy.mpl.geoaxes

    cartopy_imported = True
except (ImportError, AssertionError) as err:
    cartopy_imported = False
    
    if isinstance(err, AssertionError):
        warnings.warn(f'Cartopy not imported due to an AssertionError: {err}.')

import asilib
import asilib.map
import asilib.utils as utils


class Imager:
    """
    The central asilib class to plot, animate, and analyze ASI data.

    Normally asilib.Imager() should not be directly called by users, but by the ASI wrapper functions.
    This interface is thoroughly documented in the :ref:`contribute_asi` documentation page.

    Parameters
    ----------
    file_info: dict
        Specifies image file paths, start end end times for each file, the loader function,
        and if the user needs one or multiple images.
    meta: dict
        Specifies ASI metadata that describes the ASI name, location, cadence, and pixel resolution.
    skymap: dict
        Specifies what each pixel maps to in (azimuth, elevation) coordinates as well as
        (latitude, longitude) coordinates at a prescribed auroral mission altitude.
    plot_settings: dict
        An optional dictionary customizing the plot colormap, color scale (logarithmic vs linear),
        and color bounds (vmin, vmax arguments in matplotlib.imshow()).
    """

    def __init__(
        self,
        file_info: dict,
        meta: dict,
        skymap: dict,
        plot_settings: dict = {},
    ) -> None:
        self.file_info = {k.lower(): v for k, v in file_info.items()}
        self.meta = {k.lower(): v for k, v in meta.items()}
        self.skymap = {k.lower(): v for k, v in skymap.items()}
        self.plot_settings = {k.lower(): v for k, v in plot_settings.items()}
        self._accumulate_n = 1
        return

    def plot_fisheye(
        self,
        ax: plt.Axes = None,
        label: bool = True,
        color_map: str = None,
        color_bounds: List[float] = None,
        color_norm: str = None,
        azel_contours: bool = False,
        azel_contour_color: str = 'yellow',
        cardinal_directions: str = 'NE',
        origin: tuple = (0.8, 0.1),
    ) -> Tuple[plt.Axes, matplotlib.collections.QuadMesh]:
        """
        Plots one fisheye image, oriented with North on the top, and East on the left of the image.

        Parameters
        ----------
        ax: plt.Axes
            The subplot to plot the image on. If None this method will create one.
        label: bool
            Flag to add the "asi_array_code/location_code/image_time" text to the plot.
        color_map: str
            The matplotlib colormap to use. By default will use a black-white colormap.
            For more information See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic. The norm is not applied to RGB images (see 
            `matplotlib.pyplot.imshow <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html>`_)
        azel_contours: bool
            Superpose azimuth and elevation contours on or off.
        azel_contour_color: str
            The color of the azimuth and elevation contours.
        cardinal_directions: str
            Plot one or more cardinal directions specified with a string containing the first
            letter of one or more cardinal directions. Case insensitive. For example, to plot
            the North and East directions, set cardinal_directions='NE'.
        origin: tuple
            The origin of the cardinal direction arrows.

        Returns
        -------
        ax: plt.Axes
            The subplot object to modify the axis, labels, etc.
        im: matplotlib.collections.QuadMesh
            The plt.imshow image object. Common use for im is to add a colorbar.
            The image is oriented in the map orientation (north is up, south is down,
            west is right, and east is left). Set azel_contours=True to confirm.

        Raises
        ------
        NotImplementedError
            If the colormap is unspecified ('auto' by default) and the
            auto colormap is undefined for an ASI array.
        ValueError
            If the color_norm kwarg is not "log" or "lin".

        Example
        -------
        >>> # A bright auroral arc that was analyzed by Imajo et al., 2021 "Active
        >>> # auroral arc powered by accelerated electrons from very high altitudes"
        >>> from datetime import datetime
        >>> import matplotlib.pyplot as plt
        >>> import asilib.asi
        >>>
        >>> asi = asilib.asi.themis('RANK', time=datetime(2017, 9, 15, 2, 34, 0))
        >>> ax, im = asi.plot_fisheye(cardinal_directions='NE', origin=(0.95, 0.05))
        >>> plt.colorbar(im)
        >>> ax.axis('off')
        >>> plt.show()
        """
        if ax is None:
            _, ax = plt.subplots()

        self_copy = self.__getitem__(self.file_info['time'])
        time, image = self_copy.data

        color_map, color_norm = self._plot_params(image, color_bounds, color_map, color_norm)

        if len(self.meta['resolution']) == 3:  # tests if rgb
            vmin, vmax = (color_norm.vmin, color_norm.vmax)
            image = self._rgb_replacer(image)
            image = utils.stretch_contrast(image, vmin, vmax)

        if isinstance(color_norm, matplotlib.colors.LogNorm):
            # Increase the corner pixels with 0 counts to 1 count so 
            # it shows up black in log-scale.
            image[np.where(np.isnan(image))] = 1
        im = ax.imshow(image, cmap=color_map, norm=color_norm, origin="lower")
        if label:
            self._add_fisheye_label(time, ax)
        if azel_contours:
            self._add_azel_contours(ax, color=azel_contour_color)
        if cardinal_directions is not None:
            self._add_cardinal_directions(ax, cardinal_directions, origin=origin)
        return ax, im

    def animate_fisheye(self, **kwargs) -> None:
        """
        A wrapper for the ```Imager.animate_fisheye_gen()``` method that animates a series of
        fisheye images. Any kwargs are passed directly into ```Imager.animate_fisheye_gen()```.

        Parameters
        ----------
        ax: plt.Axes
            The optional subplot that will be drawn on.
        label: bool
            Flag to add the "asi_array_code/location_code/image_time" text to the plot.
        color_map: str
            The matplotlib colormap to use. By default will use a black-white colormap.
            For more information See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic. The norm is not applied to RGB images (see 
            `matplotlib.pyplot.imshow <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html>`_)
        azel_contours: bool
            Superpose azimuth and elevation contours on or off.
        azel_contour_color: str
            The color of the azimuth and elevation contours.
        cardinal_directions: str
            Plot one or more cardinal directions specified with a string containing the first
            letter of one or more cardinal directions. Case insensitive. For example, to plot
            the North and East directions, set cardinal_directions='NE'.
        origin: tuple
            The origin of the cardinal direction arrows.
        movie_container: str
            The movie container: mp4 has better compression but avi was determined
            to be the official container for preserving digital video by the
            National Archives and Records Administration.
        ffmpeg_params: dict
            The additional/overwitten ffmpeg output parameters. The default parameters are:
            framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
        overwrite: bool
            If true, the output will be overwritten automatically. If false it will
            prompt the user to answer y/n.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            If the colormap is unspecified ('auto' by default) and the
            auto colormap is undefined for an ASI array.
        ValueError
            If the color_norm kwarg is not "log" or "lin".

        Example
        -------
        >>> from datetime import datetime
        >>> import asilib.asi
        >>>
        >>> time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
        >>> asi = asilib.asi.themis('FSMI', time_range=time_range)
        >>> asi.animate_fisheye(cardinal_directions='NE', origin=(0.95, 0.05), overwrite=True)
        >>> print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')
        """
        movie_generator = self.animate_fisheye_gen(**kwargs)

        for _ in movie_generator:
            pass
        return

    def animate_fisheye_gen(
        self,
        ax: plt.Axes = None,
        label: bool = True,
        color_map: str = None,
        color_bounds: List[float] = None,
        color_norm: str = None,
        azel_contours: bool = False,
        azel_contour_color: str = 'yellow',
        cardinal_directions: str = 'NE',
        origin: tuple = (0.8, 0.1),
        movie_container: str = 'mp4',
        animation_save_dir: Union[pathlib.Path, str]=None,
        ffmpeg_params={},
        overwrite: bool = False,
    ) -> Generator[
        Tuple[datetime.datetime, np.ndarray, plt.Axes, matplotlib.collections.QuadMesh], None, None
    ]:
        """
        Animate a series of fisheye images and superpose your data on each image.

        A generator behaves like an iterator in that it plots one fisheye image
        at a time and yields (similar to returns) the image. You can modify, or add
        content to the image (such as a spacecraft position). Then, once the iteration
        is complete, this method stitches the images into an animation. See the examples
        below and in the examples page for use cases. The ```animate_fisheye()``` method
        takes care of the iteration.

        Parameters
        ----------
        ax: plt.Axes
            The optional subplot that will be drawn on.
        label: bool
            Flag to add the "asi_array_code/location_code/image_time" text to the plot.
        color_map: str
            The matplotlib colormap to use. By default will use a black-white colormap.
            For more information See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic. The norm is not applied to RGB images (see 
            `matplotlib.pyplot.imshow <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html>`_)
        azel_contours: bool
            Superpose azimuth and elevation contours on or off.
        azel_contour_color: str
            The color of the azimuth and elevation contours.
        cardinal_directions: str
            Plot one or more cardinal directions specified with a string containing the first
            letter of one or more cardinal directions. Case insensitive. For example, to plot
            the North and East directions, set cardinal_directions='NE'.
        origin: tuple
            The origin of the cardinal direction arrows.
        movie_container: str
            The movie container: mp4 has better compression but avi was determined
            to be the official container for preserving digital video by the
            National Archives and Records Administration.
        ffmpeg_params: dict
            The additional/overwitten ffmpeg output parameters. The default parameters are:
            framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
        overwrite: bool
            Overwrite the animation. If False, ffmpeg will prompt you to answer y/n if the
            animation already exists.

        Yields
        ------
        image_time: datetime.datetime
            The time of the current image.
        image: np.ndarray
            A 2d image array of the image corresponding to image_time
        ax: plt.Axes
            The subplot object to modify the axis, labels, etc.
        im: matplotlib.collections.QuadMesh
            The plt.imshow image object. Common use for im is to add a colorbar.
            The image is oriented in the map orientation (north is up, south is down,
            west is right, and east is left). Set azel_contours=True to confirm.

        Raises
        ------
        NotImplementedError
            If the colormap is unspecified ('auto' by default) and the
            auto colormap is undefined for an ASI array.
        ValueError
            If the color_norm kwarg is not "log" or "lin".

        Example
        -------
        >>> from datetime import datetime
        >>> import asilib.asi
        >>> import asilib
        >>>
        >>> time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
        >>> asi = asilib.asi.themis('FSMI', time_range=time_range)
        >>> gen = asi.animate_fisheye_gen(cardinal_directions='NE', origin=(0.95, 0.05), overwrite=True)
        >>> for image_time, image, ax, im in gen:
        ...         # Add your code that modifies each image here.
        ...         pass
        ...
        >>> print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')
        """
        if ax is None:
            _, ax = plt.subplots()

        # Create the animation directory inside asilib.config['ASI_DATA_DIR'] if it does
        # not exist.
        if animation_save_dir is None:
            _path = asilib.config['ASI_DATA_DIR']
        else:
            _path = animation_save_dir
        image_save_dir = pathlib.Path(
            _path,
            'animations',
            'images',
            f'{self.file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_{self.meta["array"].lower()}_'
            f'{self.meta["location"].lower()}_fisheye',
        )
        self.animation_name = (
            f'{self.file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_'
            f'{self.file_info["time_range"][-1].strftime("%H%M%S")}_'
            f'{self.meta["array"].lower()}_{self.meta["location"].lower()}_fisheye.{movie_container}'
        )
        movie_save_path = image_save_dir.parents[1] / self.animation_name

        # If the image directory exists we need to first remove all of the images to avoid
        # animating images from different method calls.
        if image_save_dir.is_dir():
            shutil.rmtree(image_save_dir)
        image_save_dir.mkdir(parents=True)

        image_paths = []
        _progressbar = utils.progressbar(
            enumerate(self.__iter__()),
            iter_length=self._estimate_n_times(),
            text=self.animation_name,
        )

        for i, (image_time, image) in _progressbar:
            ax.clear()
            ax.axis('off')
            # Use an underscore so the original method parameters are not overwritten.
            _color_map, _color_norm = self._plot_params(image, color_bounds, color_map, color_norm)
            
            if len(self.meta['resolution']) == 3:  # tests if rgb
                vmin, vmax = (_color_norm.vmin, _color_norm.vmax)
                image = self._rgb_replacer(image)
                image = utils.stretch_contrast(image, vmin, vmax)

            if isinstance(color_norm, matplotlib.colors.LogNorm):
                # Increase the corner pixels with 0 counts to 1 count so 
                # it shows up black in log-scale.
                image[np.where(np.isnan(image))] = 1
            im = ax.imshow(image, cmap=_color_map, norm=_color_norm, origin='lower')
            if label:
                self._add_fisheye_label(image_time, ax)

            if azel_contours:
                self._add_azel_contours(ax, color=azel_contour_color)
            if cardinal_directions is not None:
                self._add_cardinal_directions(ax, cardinal_directions, origin=origin)

            # Give the user the control of the subplot, image object, and return the image time
            # so that they can manipulate the image to add, for example, the satellite track.
            yield image_time, image, ax, im

            # Save the plot before the next iteration.
            save_name = f'{str(i).zfill(6)}.png'
            plt.savefig(image_save_dir / save_name)
            image_paths.append(image_save_dir / save_name)

        self._create_animation(image_paths, movie_save_path, ffmpeg_params, overwrite)
        return

    def plot_map(
        self,
        lon_bounds: tuple = (-160, -50),
        lat_bounds: tuple = (40, 82),
        ax: Union[plt.Axes, tuple] = None,
        coast_color: str = 'k',
        land_color: str = 'g',
        ocean_color: str = 'w',
        color_map: str = None,
        color_bounds: List[float] = None,
        color_norm: str = None,
        min_elevation: float = 10,
        asi_label: bool = True,
        pcolormesh_kwargs: dict = {},
        lon_grid:np.ndarray = None,
        lat_grid:np.ndarray = None,
    ) -> Tuple[plt.Axes, matplotlib.collections.QuadMesh]:
        """
        Projects an ASI image onto a map at an altitude that is defined in the skymap calibration 
        file.

        Parameters
        ----------
        lon_bounds: tuple
            The map's longitude bounds.
        lat_bounds: tuple
            The map's latitude bounds.
        ax: plt.Axes, tuple
            The subplot to put the map on. If cartopy is installed, ```ax``` must be
            a two element tuple specifying the ``plt.Figure`` object and subplot position
            passed directly as ``args`` into
            `fig.add_subplot() <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.add_subplot>`_.
        coast_color: str
            The coast color. If None will not draw it.
        land_color: str
            The land color. If None will not draw it.
        ocean_color: str
            The ocean color. If None will not draw it.
        color_map: str
            The matplotlib colormap to use. See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic. The norm is not applied to RGB images (see 
            `matplotlib.pyplot.imshow <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html>`_)
        min_elevation: float
            Masks the pixels below min_elevation degrees.
        asi_label: bool
            Annotates the map with the ASI code in the center of the mapped image.
        color_bounds: List[float] or None
            The lower and upper values of the color scale. If None, will
            automatically set it to low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile)
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic.
        pcolormesh_kwargs: dict
            A dictionary of keyword arguments (kwargs) to pass directly into
            plt.pcolormesh.
        lon_grid:np.ndarray
            Map the image onto a custom longitude grid. Both lon_grid and lat_grid must 
            be specified.
        lat_grid:np.ndarray
            Map the image onto a custom latitude grid. Both lon_grid and lat_grid must 
            be specified.

        Returns
        -------
        plt.Axes
            The subplot object to modify the axis, labels, etc.
        matplotlib.collections.QuadMesh
            The plt.pcolormesh image object. Common use for p is to add a colorbar.

        Examples
        --------
        >>> # Project an image of STEVE onto a map.
        >>> from datetime import datetime
        >>> import matplotlib.pyplot as plt
        >>> import asilib.asi
        >>>
        >>> asi = asilib.asi.themis('ATHA', time=datetime(2010, 4, 5, 6, 7, 0))
        >>> asi.plot_map(lon_bounds=(-127, -100), lat_bounds=(45, 65))
        >>> plt.tight_layout()
        >>> plt.show()
        """
        assert 'time' in self.file_info.keys(), f'Need to specify an image time.'
        for _skymap_key in ['lat', 'lon', 'el']:
            assert _skymap_key in self.skymap.keys(), (
                f'The "{_skymap_key}" key not in ' f'Imager.skymap: got {self.skymap.keys()}'
            )

        if ax is None:
            ax = asilib.map.create_map(
                lon_bounds=lon_bounds,
                lat_bounds=lat_bounds,
                coast_color=coast_color,
                land_color=land_color,
                ocean_color=ocean_color,
            )

        self_copy = self.__getitem__(self.file_info['time'])
        _, image = self_copy.data
        color_map, color_norm = self._plot_params(image, color_bounds, color_map, color_norm)

        ax, p, _ = self._plot_mapped_image(
            ax, image, min_elevation, color_map, color_norm, asi_label, 
            pcolormesh_kwargs, lon_grid=lon_grid, lat_grid=lat_grid
        )
        return ax, p

    def _plot_mapped_image(
        self, ax, image, min_elevation, color_map, color_norm, asi_label, 
        pcolormesh_kwargs, lon_grid=None, lat_grid=None
    ):
        """
        Plot the image onto a geographic map using the modified version of plt.pcolormesh.
        """
        if (lon_grid is None) and (lat_grid is None):
            _skymap_cleaner = Skymap_Cleaner(
                self.skymap['lon'], 
                self.skymap['lat'], 
                self.skymap['el'], 
            )
            _skymap_cleaner.mask_elevation(min_elevation)
            _cleaned_lon_grid, _cleaned_lat_grid = _skymap_cleaner.remove_nans()
        else:
            _cleaned_lon_grid, _cleaned_lat_grid = lon_grid, lat_grid


        if len(self.meta['resolution']) == 3:
            vmin, vmax = self.get_color_bounds()
            image = self._rgb_replacer(image)
            image = utils.stretch_contrast(image, vmin, vmax)

        pcolormesh_kwargs_copy = pcolormesh_kwargs.copy()
        if cartopy_imported and isinstance(ax, cartopy.mpl.geoaxes.GeoAxes):
            assert 'transform' not in pcolormesh_kwargs.keys(), (
                f"The pcolormesh_kwargs in Imager.plot_map() can't contain "
                f"'transform' key because it is reserved for cartopy."
            )
            pcolormesh_kwargs_copy['transform'] = ccrs.PlateCarree()
        pcolormesh_kwargs_copy['norm'] = color_norm
        pcolormesh_kwargs_copy['cmap'] = color_map
        p = ax.pcolormesh(_cleaned_lon_grid, _cleaned_lat_grid, image, **pcolormesh_kwargs_copy)

        if asi_label:
            if cartopy_imported and isinstance(ax, cartopy.mpl.geoaxes.GeoAxes):
                transform = ccrs.PlateCarree()
            else:
                transform = ax.transData
            label = ax.text(
                self.meta['lon'],
                self.meta['lat'],
                self.meta['location'].upper(),
                color='r',
                transform=transform,
                va='center',
                ha='center',
            )
        else:
            label = None
        return ax, p, label

    def animate_map(self, **kwargs) -> None:
        """
        A wrapper for the ```animate_map_gen()``` method that animates a series of
        mapped images. Any kwargs are passed directly into ```animate_map_gen()```.

        Parameters
        ----------
        ax: plt.Axes
            The optional subplot that will be drawn on.
        timestamp: bool
            If True, the time stamp is annotated in the upper left corner of the map.
        label: bool
            Flag to add the "asi_array_code/location_code/image_time" text to the plot.
        color_map: str
            The matplotlib colormap to use. By default will use a black-white colormap.
            For more information See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic. The norm is not applied to RGB images (see 
            `matplotlib.pyplot.imshow <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html>`_)
        azel_contours: bool
            Superpose azimuth and elevation contours on or off.
        azel_contour_color: str
            The color of the azimuth and elevation contours.
        cardinal_directions: str
            Plot one or more cardinal directions specified with a string containing the first
            letter of one or more cardinal directions. Case insensitive. For example, to plot
            the North and East directions, set cardinal_directions='NE'.
        movie_container: str
            The movie container: mp4 has better compression but avi was determined
            to be the official container for preserving digital video by the
            National Archives and Records Administration.
        ffmpeg_params: dict
            The additional/overwitten ffmpeg output parameters. The default parameters are:
            framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
        overwrite: bool
            If true, the output will be overwritten automatically. If false it will
            prompt the user to answer y/n.

        Example
        -------
        .. code-block:: python

            >>> from datetime import datetime
            >>> import asilib.asi
            >>> import asilib
            >>>
            >>> location = 'FSMI'
            >>> time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
            >>> asi = asilib.asi.themis(location, time_range=time_range)
            >>> asi.animate_map(overwrite=True)
            >>> print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')

        """
        movie_generator = self.animate_map_gen(**kwargs)

        for _ in movie_generator:
            pass
        return

    def animate_map_gen(
        self,
        timestamp: bool = True,
        lon_bounds: tuple = (-160, -50),
        lat_bounds: tuple = (40, 82),
        ax: Union[plt.Axes, tuple] = None,
        coast_color: str = 'k',
        land_color: str = 'g',
        ocean_color: str = 'w',
        color_map: str = None,
        color_bounds: List[float] = None,
        color_norm: str = None,
        min_elevation: float = 10,
        pcolormesh_kwargs: dict = {},
        asi_label: bool = True,
        lon_grid:np.ndarray = None,
        lat_grid:np.ndarray = None,
        movie_container: str = 'mp4',
        animation_save_dir: Union[pathlib.Path, str]=None,
        ffmpeg_params={},
        overwrite: bool = False,
    ) -> Generator[
        Tuple[datetime.datetime, np.ndarray, plt.Axes, matplotlib.collections.QuadMesh], None, None
    ]:
        """
        Animate a series of mapped images and superpose your data on each image.

        A generator behaves like an iterator in that it plots one fisheye image
        at a time and yields (similar to returns) the image. You can modify, or add
        content to the image (such as a spacecraft position). Then, once the iteration
        is complete, this method stitches the images into an animation. See the examples
        below and in the examples page for use cases. The ```animate_fisheye()``` method
        takes care of the iteration.

        Parameters
        ----------
        timestamp: bool
            If True, the time stamp is annotated in the upper left corner of the map.
        lon_bounds: tuple
            The map's longitude bounds.
        lat_bounds: tuple
            The map's latitude bounds.
        ax: plt.Axes, tuple
            The subplot to put the map on. If cartopy is installed, ```ax``` must be
            a two element tuple specifying the ``plt.Figure`` object and subplot position
            passed directly as ``args`` into
            `fig.add_subplot() <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.add_subplot>`_.
        coast_color: str
            The coast color. If None will not draw it.
        land_color: str
            The land color. If None will not draw it.
        ocean_color: str
            The ocean color. If None will not draw it.
        color_map: str
            The matplotlib colormap to use. See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_
            for supported colormaps.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic. The norm is not applied to RGB images (see 
            `matplotlib.pyplot.imshow <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html>`_)
        min_elevation: float
            Masks the pixels below min_elevation degrees.
        pcolormesh_kwargs: dict
            A dictionary of keyword arguments (kwargs) to pass directly into
            plt.pcolormesh.
        asi_label: bool
            Annotates the map with the ASI code in the center of the mapped image.
        lon_grid:np.ndarray
            Map the image onto a custom longitude grid. Both lon_grid and lat_grid must 
            be specified.
        lat_grid:np.ndarray
            Map the image onto a custom latitude grid. Both lon_grid and lat_grid must 
            be specified.
        movie_container: str
            The movie container: mp4 has better compression but avi was determined
            to be the official container for preserving digital video by the
            National Archives and Records Administration.
        ffmpeg_params: dict
            The additional/overwitten ffmpeg output parameters. The default parameters are:
            framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
        overwrite: bool
            Overwrite the animation. If False, ffmpeg will prompt you to answer y/n if the
            animation already exists.

        Yields
        ------
        image_time: datetime.datetime
            The time of the current image.
        image: np.ndarray
            A 2d image array of the image corresponding to image_time
        ax: plt.Axes
            The subplot object to modify the axis, labels, etc.
        im: matplotlib.collections.QuadMesh
            The plt.imshow image object. Common use for im is to add a colorbar.
            The image is oriented in the map orientation (north is up, south is down,
            west is right, and east is left). Set azel_contours=True to confirm.

        Example
        -------
        .. code-block:: python

            >>> from datetime import datetime
            >>> import matplotlib.pyplot as plt
            >>> import asilib
            >>> import asilib.asi
            >>>
            >>> location = 'FSMI'
            >>> time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
            >>> asi = asilib.asi.themis(location, time_range=time_range)
            >>> ax = asilib.map.create_map(lon_bounds=(-120, -100), lat_bounds=(55, 65))
            >>> plt.tight_layout()
            >>>
            >>> gen = asi.animate_map_gen(overwrite=True, ax=ax)
            >>>
            >>> for image_time, image, ax, im in gen:
            >>>     # Add your code that modifies each image here...
            >>>     # To demonstrate, lets annotate each frame with the timestamp.
            >>>     # We will need to delete the prior text object, otherwise the current one
            >>>     # will overplot on the prior one---clean up after yourself!
            >>>     if 'text_obj' in locals():
            >>>             ax.texts.remove(text_obj)
            >>>     text_obj = ax.text(0, 0.9, f'THEMIS-{location} at {image_time:%F %T}',
            >>>             transform=ax.transAxes, color='white', fontsize=15)
            >>>
            >>> print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')
        """
        if ax is None:
            ax = asilib.map.create_map(
                lon_bounds=lon_bounds,
                lat_bounds=lat_bounds,
                coast_color=coast_color,
                land_color=land_color,
                ocean_color=ocean_color,
            )
        # Create the animation directory inside asilib.config['ASI_DATA_DIR'] if it does
        # not exist.
        if animation_save_dir is None:
            _path = asilib.config['ASI_DATA_DIR']
        else:
            _path = animation_save_dir
        image_save_dir = pathlib.Path(
            _path,
            'animations',
            'images',
            f'{self.file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_{self.meta["array"].lower()}_'
            f'{self.meta["location"].lower()}_map',
        )

        self.animation_name = (
            f'{self.file_info["time_range"][0].strftime("%Y%m%d_%H%M%S")}_'
            f'{self.file_info["time_range"][-1].strftime("%H%M%S")}_'
            f'{self.meta["array"].lower()}_{self.meta["location"].lower()}_map.{movie_container}'
        )
        movie_save_path = image_save_dir.parents[1] / self.animation_name
        # If the image directory exists we need to first remove all of the images to avoid
        # animating images produced by different method calls.
        if image_save_dir.is_dir():
            shutil.rmtree(image_save_dir)
        image_save_dir.mkdir(parents=True)

        image_paths = []
        _progressbar = utils.progressbar(
            enumerate(self.__iter__()),
            iter_length=self._estimate_n_times(),
            text=self.animation_name,
        )
        for i, (image_time, image) in _progressbar:
            # Use an underscore so the original method parameters are not overwritten.
            _color_map, _color_norm = self._plot_params(image, color_bounds, color_map, color_norm)

            ax, pcolormesh_obj, label_obj = self._plot_mapped_image(
                ax, image, min_elevation, _color_map, _color_norm, asi_label, 
                pcolormesh_kwargs, lon_grid=lon_grid, lat_grid=lat_grid
            )

            # Give the user the control of the subplot, image object, and return the image time
            # so that they can manipulate the image to add, for example, the satellite track.

            if timestamp:
                if '_text_obj' in locals():
                    _text_obj.remove() # noqa: F821
                _text_obj = ax.text(
                    0.01, 
                    0.99, 
                    f'{image_time: %Y-%m-%d %H:%M:%S}', 
                    va='top', 
                    transform=ax.transAxes,
                    fontsize=12,
                    color='white',
                    bbox=dict(facecolor='grey', edgecolor='black', alpha=0.5)
                )
            yield image_time, image, ax, pcolormesh_obj

            # Save the plot before the next iteration.
            save_name = f'{str(i).zfill(6)}.png'
            plt.savefig(image_save_dir / save_name)
            image_paths.append(image_save_dir / save_name)

            # Clean up the objects that this method generated.
            if label_obj is not None:
                # ax.texts.remove(label_obj)
                label_obj.remove()
            # ax.collections.remove(pcolormesh_obj)
            pcolormesh_obj.remove()

        self._create_animation(image_paths, movie_save_path, ffmpeg_params, overwrite)
        return

    def keogram(
        self, path: np.array = None, aacgm=False, minimum_elevation: float = 0
    ) -> Tuple[np.array, np.array, np.array]:
        """
        Create a keogram along the meridian or a custom path.

        Parameters
        ----------
        path: array
            Make a keogram along a custom path. The path is a lat/lon array of shape (n, 2).
            Longitude must be between [-180, 180].
        aacgm: bool
            Map the keogram latitudes to Altitude Adjusted Corrected Geogmagnetic Coordinates
            (aacgmv2) derived by Shepherd, S. G. (2014), Altitude-adjusted corrected geomagnetic
            coordinates: Definition and functional approximations, Journal of Geophysical
            Research: Space Physics, 119, 7501-7521, doi:10.1002/2014JA020264.
        minimum_elevation: float
            The minimum elevation of pixels to use in the keogram.

        Returns
        -------
        np.array:
            Keogram timestamps
        np.array:
            Keogram latitudes: geographic if the aacgm kwarg is False and magnetic is
            aacgm if True.
        np.array:
            Keogram array with rows corresponding to times and columns with latitudes.

        Example
        -------
        >>> # A keogram in geographic and magnetic latitude coordinates. See
        >>> # Imager.plot_keogram() on how to plot a keogram.
        >>> # Event from https://doi.org/10.1029/2021GL094696
        >>> import numpy as np
        >>> import asilib.asi
        >>>
        >>> time_range=['2008-01-16T10', '2008-01-16T12']
        >>>
        >>> asi = asilib.asi.themis('GILL', time_range=time_range)
        >>> time, geo_lat, geo_keogram = asi.keogram()  # geographic latitude
        >>> time, mag_lat, mag_keogram = asi.keogram(aacgm=True)  # magnetic latitude
        >>> time
        array([datetime.datetime(2008, 1, 16, 10, 0, 0, 20162),
            datetime.datetime(2008, 1, 16, 10, 0, 3, 9658),
            datetime.datetime(2008, 1, 16, 10, 0, 6, 29345), ...,
            datetime.datetime(2008, 1, 16, 11, 59, 51, 50496),
            datetime.datetime(2008, 1, 16, 11, 59, 54, 10602),
            datetime.datetime(2008, 1, 16, 11, 59, 57, 60543)], dtype=object)
        >>> geo_lat[:10]
        array([47.900368, 48.506763, 49.057587, 49.556927, 50.009083, 50.418365,
            50.788963, 51.124836, 51.42965 , 51.706768], dtype=float32)
        >>> mag_lat[:10]
        array([57.97198543, 58.55886486, 59.09144098, 59.57379565, 60.01019679,
            60.40490277, 60.76203547, 61.0854798 , 61.37882613, 61.64536043])
        >>> np.all(mag_keogram == geo_keogram)  # aacgm=True only changes the latitudes.
        True
        """
        # Determine what pixels to slice (path or along the meridian).
        self._keogram_pixels(path, minimum_elevation)

        # Allocate the keogram time and image arrays. Currently the black and white (B&W)
        # and color (RGB) images are allocated separately. In the future we should try to
        # consolidate this into one self._keogram array allocation.
        self._keogram_time = np.nan * np.zeros(self._estimate_n_times(), dtype=object)
        if len(self.meta['resolution']) == 2:
            # B&W images.
            self._keogram = np.nan * np.zeros((self._estimate_n_times(), self._pixels.shape[0]))
        elif len(self.meta['resolution']) == 3:
            # RGB images.
            self._keogram = np.nan * np.zeros(
                (self._estimate_n_times(), self._pixels.shape[0], self.meta['resolution'][2])
            )

        self._geogram_lat = self._keogram_latitude(aacgm)

        if (hasattr(self, '_times')) and (hasattr(self, '_images')):
            # if self.data() was already called
            self._keogram[0 : self._images.shape[0], ...] = self._images[
                :, self._pixels[:, 0], self._pixels[:, 1], ...
            ]
            self._keogram_time = self._times
        else:
            # Otherwise load the data, one file at a time.
            start_time_index = 0
            _progressbar = utils.progressbar(
                self.iter_files(),
                iter_length=np.array(self.file_info['path']).shape[0],
                text=f'{self.meta["array"]} {self.meta["location"]} keogram',
            )
            for file_times, file_images in _progressbar:
                end_time_index = start_time_index + file_images.shape[0]

                self._keogram[start_time_index:end_time_index, ...] = file_images[
                    :, self._pixels[:, 0], self._pixels[:, 1], ...
                ]
                self._keogram_time[start_time_index:end_time_index] = file_times
                start_time_index += file_images.shape[0]

        # Remove NaN keogram rows (unfilled or the data is NaN.).
        i_valid = np.where(~np.isnan(self._keogram[:, 0, ...]))[0]
        self._keogram = self._keogram[i_valid, ...]
        self._keogram_time = self._keogram_time[i_valid]

        if self._keogram.shape[0] == 0:
            raise ValueError(
                f"The keogram is empty for {self.meta['array']}/{self.meta['location']} "
                f"during {self.file_info['time_range']}. The images likely don't exist "
                f"in this time interval."
            )
        return self._keogram_time, self._geogram_lat, self._keogram

    def plot_keogram(
        self,
        ax: plt.Axes = None,
        path: np.array = None,
        aacgm=False,
        title: bool = True,
        minimum_elevation: float = 0,
        color_map: str = None,
        color_bounds: List[float] = None,
        color_norm: str = None,
        pcolormesh_kwargs={},
    ) -> Tuple[plt.Axes, matplotlib.collections.QuadMesh]:
        """
        Plot a keogram along the meridian or a custom path.

        Parameters
        ----------
        ax: plt.Axes
            The subplot to plot the keogram on.
        path: np.array
            Make a keogram along a custom path. The path is a lat/lon array of shape (n, 2).
            Longitude must be between [-180, 180].
        aacgm: bool
            Map the keogram latitudes to Altitude Adjusted Corrected Geogmagnetic Coordinates
            (aacgmv2) derived by Shepherd, S. G. (2014), Altitude-adjusted corrected geomagnetic
            coordinates: Definition and functional approximations, Journal of Geophysical
            Research: Space Physics, 119, 7501-7521, doi:10.1002/2014JA020264.
        title: bool
            Add a plot title with the date, ASI array, and ASI location.
        minimum_elevation: float
            The minimum elevation of pixels to use in the keogram.
        color_map: str
            The matplotlib colormap to use. See `matplotlib colormaps <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_
            for supported colormaps.
        color_bounds: List[float]
            The lower and upper values of the color scale. The default is: low=1st_quartile and
            high=min(3rd_quartile, 10*1st_quartile). This range works well for most cases.
        color_norm: str
            Set the 'lin' (linear) or 'log' (logarithmic) color normalization. If color_norm=None,
            the color normalization will be taken from the ASI array (if specified), and if not
            specified it will default to logarithmic. The norm is not applied to RGB images (see 
            `matplotlib.pyplot.imshow <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html>`_)
        pcolormesh_kwargs: dict
            A dictionary of keyword arguments (kwargs) to pass directly into
            plt.pcolormesh.

        Returns
        -------
        plt.Axes
            The subplot object to modify the axis, labels, etc.
        matplotlib.collections.QuadMesh
            The plt.pcolormesh image object, useful to add a colorbar, for example.

        Example
        -------
        >>> # Plot a keogram in geographic and magnetic latitude coordinates.
        >>> # Event from https://doi.org/10.1029/2021GL094696
        >>> import matplotlib.pyplot as plt
        >>> import asilib.asi
        >>>
        >>> time_range=['2008-01-16T10', '2008-01-16T12']
        >>> fig, ax = plt.subplots(2, sharex=True, figsize=(10, 6))
        >>>
        >>> asi = asilib.asi.themis('GILL', time_range=time_range)
        >>> _, p = asi.plot_keogram(ax=ax[0], color_map='turbo')
        >>> asi.plot_keogram(ax=ax[1], color_map='turbo', aacgm=True, title=False)
        >>>
        >>> ax[0].set_ylabel('Geographic Lat [deg]')
        >>> ax[1].set_ylabel('Magnetic Lat [deg]')
        >>> fig.subplots_adjust(right=0.8)
        >>> cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        >>> fig.colorbar(p, cax=cbar_ax)
        >>> plt.show()
        """
        if ax is None:
            _, ax = plt.subplots()

        if ('cmap' in pcolormesh_kwargs.keys()) and (color_map is not None):
            raise ValueError(
                'The colormap is defined in both the color_map and pcolormesh_kwargs["cmap"].'
            )
        if ('norm' in pcolormesh_kwargs.keys()) and (color_norm is not None):
            raise ValueError(
                'The color norm is defined in both the color_norm and pcolormesh_kwargs["norm"].'
            )

        _keogram_time, _geogram_lat, _keogram = self.keogram(
            path=path, aacgm=aacgm, minimum_elevation=minimum_elevation
        )

        _color_map, _color_norm = self._plot_params(_keogram, color_bounds, color_map, color_norm)

        _keogram = np.moveaxis(_keogram, 0, 1)  # Transpose 0th and 1st axes for pcolormesh. 
        if len(self.meta['resolution']) == 3:  # tests if rgb
            _keogram = self._rgb_replacer(_keogram)
            vmin, vmax = self.get_color_bounds()
            _keogram = utils.stretch_contrast(_keogram, vmin, vmax)            

        pcolormesh_obj = ax.pcolormesh(
            _keogram_time,
            _geogram_lat,
            _keogram,
            norm=_color_norm,
            shading='auto',
            cmap=_color_map,
            **pcolormesh_kwargs,
        )

        if title:
            ax.set_title(
                f'{self.file_info["time_range"][0].date()} | {self.meta["array"]}-{self.meta["location"]} keogram'
            )
        return ax, pcolormesh_obj
    
    def set_color_bounds(self, lower, upper):
        """
        Sets the default color bounds for all subsequent calls to plotting functions, unless
        later overwritten by the color_bounds kawarg in each plotting/animating method. 

        Parameters
        ----------
        lower, upper: float
            The lower and upper color limits. 
        """
        self.plot_settings['color_bounds'] = (lower, upper)
        return
    
    def get_color_bounds(self):
        """
        Get the default (vmin, vmax) color bounds for the imager.
        """
        return self.plot_settings['color_bounds']
    
    def auto_color_bounds(self, images:np.ndarray=None):
        """
        Calculate the (vmin, vmax) color bounds automatically by loading in a subset of the 
        image data.

        Parameters
        ----------
        images: np.ndarray
            If None, will calculate the bounds range automatically by loading in a subset of the 
            image data, or return the pre-defined color bounds in the imager.plot_settings dict.
            If a np.array, will calculate the color bounds directly from the images.
        """
        if images is not None:
            lower, upper = np.quantile(images, (0.25, 0.98))
            return [lower, np.min([upper, lower * 10])]
        
        num = min(len(self.file_info['start_time']), 3)
        file_indicies = np.arange(
            0, 
            len(self.file_info['start_time']), 
            len(self.file_info['start_time'])//num
            ).astype(int)
        flattened_images = np.array([])

        for path in np.array(self.file_info['path'])[file_indicies]:
            _, _file_images = self.file_info['loader'](path)
            if len(self.meta['resolution']) == 3:
                _file_images = self._rgb_replacer(_file_images)
            flattened_images = np.append(flattened_images, _file_images.flatten())
        
        # You get nans if we're working with a subset of the RGB channels.
        valid_idx = np.where(~np.isnan(flattened_images))[0]
        lower, upper = np.quantile(flattened_images[valid_idx], (0.25, 0.98))
        return lower, np.min([upper, lower * 10])
    
    def _keogram_pixels(self, path, minimum_elevation=20):
        """
        Find what pixels to index and reshape the keogram.
        """
        # CASE 1: No path provided. Output self._pixels that slice the meridian.
        if path is None:
            self._pixels = -1*np.ones((self.meta['resolution'][0], 2), dtype=int)
            self._pixels[:, 0] = np.arange(self.meta['resolution'][0])
            self._pixels[:, 1] = int(self.meta['resolution'][1] // 2)

        # CASE 2: A path is provided so now we need to calculate the custom path
        # on the lat/lon skymap.
        else:
            self._pixels = self._path_to_pixels(path)

        above_elevation = np.where(
            (self.skymap['el'][self._pixels[:, 0], self._pixels[:, 1]] >= minimum_elevation)
            & (np.isfinite(self.skymap['lat'][self._pixels[:, 0], self._pixels[:, 1]]))
        )[0]
        self._pixels = self._pixels[above_elevation]
        return

    def _path_to_pixels(self, path: np.array, threshold: float = 1) -> np.array:
        """
        Map a lat/lon path to ASI x- and y-pixels.

        Parameters
        ----------
        path: np.array
            The lat/lon array of shape (n, 2). Longitude must be between [-180, 180]
        threshold: float
            The maximum distance threshold, in degrees, between the path (lat, lon)
            and the skymap (lat, lon)

        Returns
        -------
        np.array
            ASI x and y indices along the path.
        """
        if np.array(path).shape[0] == 0:
            raise ValueError('The path is empty.')
        if np.any(np.isnan(path)):
            raise ValueError("The lat/lon path can't contain NaNs.")
        if np.any(np.max(path) > 180) or np.any(np.min(path) < -180):
            raise ValueError("The lat/lon values must be in the range -180 to 180.")

        nearest_pixels = np.nan * np.zeros_like(path)

        for i, (lat, lon) in enumerate(path):
            distances = _haversine(
                self.skymap['lat'],
                self.skymap['lon'],
                lat * np.ones_like(self.skymap['lon']),
                lon * np.ones_like(self.skymap['lon']),
            )
            idx = np.where(distances == np.nanmin(distances))

            if distances[idx][0] > threshold:
                warnings.warn(
                    f'Some of the keogram path coordinates are outside of the '
                    f'maximum {threshold} degrees distance from the nearest '
                    f'skymap map coordinate.'
                )
                continue
            nearest_pixels[i, :] = [idx[0][0], idx[1][0]]

        valid_pixels = np.where(np.isfinite(nearest_pixels[:, 0]))[0]
        if valid_pixels.shape[0] == 0:
            raise ValueError('The keogram path is completely outside of the skymap.')
        output = nearest_pixels[valid_pixels, :].astype(int)
        # If the image.shape[0] = lat_skymap.shape[0]-1 as for the Calgary's skymaps
        # that are defined at pixel edges, we need to check and remove nearest_pixels
        # that equal self.meta['resolution'][0].
        if self.meta['resolution'][0] == self.skymap['lat'].shape[0]-1:
            output[np.where(output[:, 0]==self.meta['resolution'][0])[0], 0] -= 1
        return output

    def _keogram_latitude(self, aacgm):
        """
        Keogram's vertical axis: geographic latitude, magnetic latitude, or pixel index.
        """
        _geo_lats = self.skymap['lat'][self._pixels[:, 0], self._pixels[:, 1]]
        assert np.all(np.isfinite(_geo_lats)), f'Some keogram lats are NaNs {_geo_lats=}.'
        if aacgm:
            _geo_lons = self.skymap['lon'][self._pixels[:, 0], self._pixels[:, 1]]
            _aacgm_lats = aacgmv2.convert_latlon_arr(
                _geo_lats,
                _geo_lons,
                self.skymap['alt'],
                self.file_info['time_range'][0],
                method_code="G2A",
            )[0]
            return _aacgm_lats
        else:
            return _geo_lats

    def __getitem__(self, _slice):
        """
        Slice an Imager object by time.

        Parameters
        ----------
        _slice: str, pd.Timestamp, datetime.datetime, or list.
            The time(s) to slice an Imager object. Can type of slice can be either
            1) [start_time:end_time], or 2) just time.

        Yields
        ------
        asilib.Imager:
            A sliced version of the Imager.
        """
        start_time, end_time = self._convert_slice(_slice)
        idx = np.where(
            (start_time <= np.array(self.file_info['end_time']))
            & (end_time >= np.array(self.file_info['start_time']))
        )[0]

        if len(idx) == 0:
            raise FileNotFoundError(f'Imager does not have any data contained in slice={_slice}')

        # Create the new variables.
        new_file_info = {}
        if start_time == end_time:
            # A single time stamp
            new_file_info['time'] = start_time
        else:
            new_file_info['time_range'] = [start_time, end_time]
        new_file_info['start_time'] = np.array(self.file_info['start_time'])[idx]
        new_file_info['end_time'] = np.array(self.file_info['end_time'])[idx]
        new_file_info['path'] = np.array(self.file_info['path'])[idx]
        new_file_info['loader'] = self.file_info['loader']

        new_meta = copy.copy(self.meta)
        new_skymap = copy.copy(self.skymap)
        new_plot_settings = copy.copy(self.plot_settings)

        cls = type(self)
        return cls(new_file_info, new_meta, new_skymap, plot_settings=new_plot_settings)

    def _convert_slice(self, _slice):
        """
        Validate and convert the slice into datetime objects.
        """
        # Convert the [start_time:end_time] slice to datetime objects.
        if isinstance(_slice, slice):
            # Check the start slice and if it is None assign time_range[0]
            if isinstance(_slice.start, str):
                start_time = dateutil.parser.parse(_slice.start)
            elif isinstance(_slice.start, (datetime.datetime, pd.Timestamp)):
                start_time = _slice.start
            elif _slice.start is None:
                start_time = self.file_info['time_range'][0]
            else:
                raise ValueError(
                    f'The start index can only be a time object, string, or None. '
                    f'{_slice.start} is unsupported'
                )

            # Check the end slice and if it is None assign time_range[1]
            if isinstance(_slice.stop, str):
                end_time = dateutil.parser.parse(_slice.stop)
            elif isinstance(_slice.stop, (datetime.datetime, pd.Timestamp)):
                end_time = _slice.stop
            elif _slice.stop is None:
                end_time = self.file_info['time_range'][1]
            else:
                raise ValueError(
                    f'The start index can only be a time object, string, or None. '
                    f'{_slice.stop} is unsupported'
                )

            if _slice.step is not None:
                raise NotImplementedError
            return start_time, end_time

        # Convert the [time] slice to datetime object.
        elif isinstance(_slice, (str, datetime.datetime, pd.Timestamp)):
            if isinstance(_slice, str):
                slice_time = dateutil.parser.parse(_slice)
            else:
                slice_time = _slice
            return slice_time, slice_time
        elif isinstance(_slice, tuple):
            raise NotImplementedError(
                'At this time asilib.Imager does not support multi-dimensional indexing.'
            )
        else:
            raise ValueError(f'The slice must be [time] or [start_time:end_time], not {_slice}.')

    def __iter__(self):
        """
        Iterate over individual timestamps and images using the
        ```for time, image in asilib.Imager(...)```.

        Parameters
        ----------
        None

        Yields
        ------
        datetime.datetime
            timestamp
        np.array
            image.
        """
        if hasattr(self, '_times') and hasattr(self, '_images'):
            for time_i, image_i in zip(self._times, self._images):
                yield time_i, image_i
        else:
            for file_times, file_images in self.iter_files():
                for time_i, image_i in zip(file_times, file_images):
                    yield time_i, image_i

    def iter_files(self) -> Union[np.array, np.array]:
        """
        Iterate one ASI file (or large chunks of a file) at a time.
        The output data is clipped by time_range.

        Yields
        ------
        np.array:
            ASI timestamps in datetine.datetime() or numpy.datetime64() format.
        np.array:
            ASI images.

        Example
        -------
        Loop over ~5 minutes of data, starting in the middle of one file.
            .. code-block:: python

                >>> import asilib.asi
                >>> time_range=['2008-01-16T10:00:30', '2008-01-16T10:05']
                >>> asi = asilib.asi.themis('GILL', time_range=time_range)
                >>> for file_times, file_images in asi.iter_files():
                ...     print(file_times[0], file_times[-1], file_images.shape)
                ...
                2008-01-16 10:00:30.019526 2008-01-16 10:00:57.049007 (10, 256, 256)
                2008-01-16 10:01:00.058996 2008-01-16 10:01:57.049620 (20, 256, 256)
                2008-01-16 10:02:00.059597 2008-01-16 10:02:57.029981 (20, 256, 256)
                2008-01-16 10:03:00.050822 2008-01-16 10:03:57.020254 (20, 256, 256)
                2008-01-16 10:04:00.030448 2008-01-16 10:04:57.046170 (20, 256, 256)
        """
        self._loader_is_gen = inspect.isgeneratorfunction(self.file_info['loader'])
        if 'time_range' not in self.file_info.keys():
            raise KeyError('Imager was not instantiated with a time_range.')
        
        self._file_times_shapes = []

        for path in self.file_info['path']:
            # Check if the loader function is a generator. If not, asilib
            # will load one image file at a time and assume that opening one file
            # won't overwhelm the PC's memory. If loader is a generator,
            # on the other hand, we need to loop over every chunk of data
            # yielded by the generator and over the timestamps in each chunk.
            if not self._loader_is_gen:
                times, images = self.file_info['loader'](path)
                self._file_times_shapes.append(times.shape[0])

                idt = np.where(
                    (times >= self.file_info['time_range'][0])
                    & (times <= self.file_info['time_range'][1])
                )[0]
                yield times[idt], images[idt]

            else:
                gen = self.file_info['loader'](path)

                for time_chunk, image_chunk in gen:
                    idt = np.where(
                        (time_chunk >= self.file_info['time_range'][0])
                        & (time_chunk <= self.file_info['time_range'][1])
                    )[0]
                    yield time_chunk[idt], image_chunk[idt]
        return

    def _estimate_n_times(self):
        """
        Estimate the maximum number of time stamps for the Imager's time range.
        """
        n_sec = (self.file_info['time_range'][1] - self.file_info['time_range'][0]).total_seconds()
        # +2 is for when time_range includes the start and end time stamps.
        # This will be trimmed later.
        return int(n_sec / self.meta['cadence']) + 2

    @property
    def data(self):
        """
        Load ASI data.

        Returns
        -------
        namedtuple
            A named tuple containing (times, images). Members can be accessed using either
            index notation, or dot notation.
        """
        _img_data_type = namedtuple('data', ['time', 'image'])

        if 'time_range' in self.file_info.keys():
            # If already run.
            if hasattr(self, '_times') and hasattr(self, '_images'):
                return _img_data_type(self._times, self._images)

            self._times = np.nan * np.zeros(self._estimate_n_times(), dtype=object)
            self._images = np.nan * np.zeros((self._estimate_n_times(), *self.meta['resolution']))

            start_idt = 0
            for file_times, file_images in self.iter_files():
                try:
                    self._times[start_idt : start_idt + file_times.shape[0]] = file_times
                except ValueError as err:
                    if 'could not broadcast input array from shape' in str(err):
                        raise ValueError(
                            f'The imager tried to load more images than expected. This is likely '
                            f'due to an inconsistent number of time stamps in the image files. To'
                            f' confirm, here are the unique image files shapes: '
                            f'{set(self._file_times_shapes)}. There should be only one unique shape.'
                        ) from err
                self._images[start_idt : start_idt + file_times.shape[0]] = file_images
                start_idt += file_times.shape[0]
            # Remove any unfilled times and images
            self._times = self._times[:start_idt]
            self._images = self._images[:start_idt, ...].astype(int)
            return _img_data_type(self._times, self._images)

        elif 'time' in self.file_info.keys():
            return _img_data_type(*self._load_image(self.file_info['time']))

        else:
            raise ValueError(
                f'This imager instance does not contain either the '
                f'"time" or "time_range" data variables. The data '
                f'variables are {self.file_info.keys()}.'
            )

    def _load_image(self, time):
        """
        Load a single image and time stamp nearest to time.

        Parameters
        ----------
        time: datetime.datetime
            The requested image time.

        Returns
        -------
        datetime.datetime
            Image timestamp
        np.array
            Image.

        Raises
        ------
        IndexError
            If the nearest image timestamp is more than self.meta['cadence'] away
            from time.
        """
        # Case where the loader is a function
        if not inspect.isgeneratorfunction(self.file_info['loader']):
            _times, _images = self.file_info['loader'](self.file_info['path'][0])
            image_index = np.argmin(np.abs([(time - t_i).total_seconds() for t_i in _times]))
            if np.abs((time - _times[image_index]).total_seconds()) > self.meta['cadence']:
                raise IndexError(
                    f'Cannot find a time stamp within {self.meta["cadence"]} seconds of '
                    f'{time}. Closest time stamp is {_times[image_index]}.'
                )
            return (
                _times[image_index],
                _images[image_index, ...],
            )  # Ellipses to load all other dimenstions.
        # Case where the loader is a generator function.
        else:
            gen = self.file_info['loader'](self.file_info['path'][0])
            for _times, _images in gen:
                image_index = np.argmin(np.abs([(time - t_i).total_seconds() for t_i in _times]))
                if np.abs((time - _times[image_index]).total_seconds()) < self.meta['cadence']:
                    return _times[image_index], _images[image_index, ...]

            raise IndexError(
                f'Cannot find a time stamp within {self.meta["cadence"]} seconds of ' f'{time}.'
            )

    def __str__(self) -> str:
        if ('time' in self.file_info.keys()) and (self.file_info['time'] is not None):
            s = (
                f'A {self.meta["array"]}-{self.meta["location"]} Imager. '
                f'time={self.file_info["time"]}'
            )
        elif ('time_range' in self.file_info.keys()) and (self.file_info['time_range'] is not None):
            s = (
                f'A {self.meta["array"]}-{self.meta["location"]} Imager. '
                f'time_range={self.file_info["time_range"]}'
            )
        return s

    def _plot_params(self, image, color_bounds, color_map, color_norm):
        """
        Sets the plot color bounds, map style, and normalization values. In order of precedence,
        these kwargs are checked first, and then self.plot_settings. If these values are
        unset in either place, defaults are returned.
        """
        # Check the plot_settings dict and fall back to a default if user did not specify
        # color_bounds in the method call.
        if color_bounds is None:            
            color_bounds = self.get_color_bounds()
        else:
            if callable(color_bounds):  # function that ouputs vmin, vmax
                color_bounds = color_bounds(image)
            else:
                color_bounds = color_bounds  # vmin, vmax

        if color_map is None:
            if 'color_map' in self.plot_settings.keys():
                color_map = self.plot_settings['color_map']
            else:
                color_map = 'Greys_r'
        else:
            color_map = color_map

        # If color_norm is specified by the method, it overrules the ASI array settings
        # or Imager setting. If the method's color_norm is not specified, check if the
        # ASI has it specified in self.plot_settings. Lastly, if the ASI does not have
        # it specified, default to a logarithmic color_norm.
        if color_norm is None:
            if 'color_norm' in self.plot_settings.keys():
                color_norm = self.plot_settings['color_norm']
            else:
                color_norm = 'log'
        if color_norm == 'log':
            color_norm = colors.LogNorm(vmin=color_bounds[0], vmax=color_bounds[1])
        elif color_norm == 'lin':
            color_norm = colors.Normalize(vmin=color_bounds[0], vmax=color_bounds[1])
        elif isinstance(color_norm, matplotlib.colors.Normalize):
            pass
        else:
            raise ValueError(f'color_norm must be either None, "log", or "lin", not {color_norm=}.')

        return color_map, color_norm

    def _add_azel_contours(self, ax: plt.Axes, color: str = 'yellow') -> None:
        """
        Adds contours of azimuth and elevation to the movie image.

        Parameters
        ----------
        ax: plt.Axes
            The subplot to draw on.
        color: str (optional)
            The contour color.
        """
        az_contours = ax.contour(
            self.skymap['az'],
            colors=color,
            linestyles='dotted',
            levels=np.arange(0, 361, 90),
            alpha=1,
        )
        el_contours = ax.contour(
            self.skymap['el'],
            colors=color,
            linestyles='dotted',
            levels=np.arange(1, 91, 30),
            alpha=1,
        )
        plt.clabel(az_contours, inline=True, fontsize=12)
        plt.clabel(el_contours, inline=True, fontsize=12, rightside_up=True)
        return

    def _add_fisheye_label(self, image_time, ax: plt.Axes):
        """
        Adds a ASI array-location-time label to the image
        """
        if self.meta['cadence'] > 1:
            time_label = image_time.strftime('%Y-%m-%d %T')
        else:
            time_label = image_time.strftime('%Y-%m-%d %T:%f')
        ax.text(
            0,
            0,
            (f"{self.meta['array'].upper()}/{self.meta['location'].upper()}\n" f"{time_label}"),
            va='bottom',
            transform=ax.transAxes,
            color='white',
            fontsize=self.plot_settings.get('label_fontsize', None)
        )
        return

    def _add_cardinal_directions(self, ax, directions, el_step=10, origin=(0.8, 0.1), length=0.1):
        """
        Plot cardinal direction arrows. See _calc_cardinal_direction() for the algorithm.

        Parameters
        ----------
        direction: str
            The cardinal direction. Must be a string containing 'N', 'S', 'E', or 'W', or a
            combination of them. Case insensitive.
        el_step: int
            The elevation steps used to fit the cardinal direction line.
        origin: tuple
            The origin of the direction arrows.
        length: float
            The arrow length.
        """
        assert isinstance(directions, str), 'Cardinal directions must be a string.'
        for direction in directions:
            direction = direction.upper()
            if direction not in ['N', 'S', 'E', 'W']:
                raise ValueError(f'Cardinality direction "{direction}" is invalid."')
            rise, run = self._calc_cardinal_direction(direction, el_step)
            norm = length / np.sqrt(rise**2 + run**2)

            ax.annotate(
                direction,
                xy=origin,
                xytext=(origin[0] + run * norm, origin[1] + rise * norm),
                arrowprops={'arrowstyle': "<-", 'color': 'w'},
                xycoords='axes fraction',
                color='w',
                ha='center',
                va='center',
                fontsize=self.plot_settings.get('label_fontsize', None)
            )
        return

    def _calc_cardinal_direction(self, direction, el_step, validate=False):
        """
        Calculate the cardinal direction arrows.

        Each direction is calculated by:
        1. Calculate the azimuths in each 5-degree elevation step.
        2. For each elevation step, find the azimuth corresponding to that cardinal direction.
        The result is a set of points going from 0 to 90 degree elevation along the cardinal
        direction.
        3. Calculate the rise and run between the two points that are nearest and furthest from
        zenith, located on the caridinal direction.

        Parameters
        ----------
        direction: str
            The cardinal direction. Must be one of 'N', 'S', 'E', or 'W' (case sensitive).
        el_step: int
            The elevation steps used to fit the cardinal direction line.

        Returns
        -------
        float
            The rise of the difference in the two pixels that are closest and furthest
            from zenith
        float
            The run of the difference in the two pixels that are closest and furthest
            from zenith
        """
        _azimuths = {'N': 0, 'E': 90, 'S': 180, 'W': 270}

        # Don't recalculate the rise and run if they've already been calculated.
        if not hasattr(self, '_cardinal_direction'):
            self._cardinal_direction = {}
        else:
            if direction in self._cardinal_direction.keys():
                return self._cardinal_direction[direction]

        elevation_steps = np.arange(0, 91, el_step)
        _direction_pixels = np.nan * np.zeros((elevation_steps.shape[0], 2), dtype=int)

        for i, (el_low, el_high) in enumerate(zip(elevation_steps[:-1], elevation_steps[1:])):
            id_el = np.where(~((self.skymap['el'] > el_low) & (self.skymap['el'] <= el_high)))
            _az = self.skymap['az'].copy()
            _az[id_el] = np.nan

            try:
                min_az_flat_array = np.nanargmin(np.abs(_az - _azimuths[direction]))
            except ValueError as err:
                if 'All-NaN slice encountered' == str(err):
                    continue
                raise
            _direction_pixels[i, :] = np.unravel_index(min_az_flat_array, self.skymap['az'].shape)

        _direction_pixels = _direction_pixels[~np.isnan(_direction_pixels[:, 0]), :]
        # Points near the edge scew the cardinal direction from the true direction.
        _not_near_edge = np.where(
            (_direction_pixels[:, 0] < self.meta['resolution'][0]-1) & 
            (_direction_pixels[:, 0] > 0) & 
            (_direction_pixels[:, 1] < self.meta['resolution'][1]-1) & 
            (_direction_pixels[:, 1] > 0)
            )[0]
        _direction_pixels = _direction_pixels[_not_near_edge, :]
        _direction_pixels = _direction_pixels.astype(int)

        # Calculate the pixels nerest and furthest away from zenith. This will define the rise
        # and run.
        center_pixel = np.array([self.skymap['az'].shape[0] // 2, self.skymap['az'].shape[1] // 2])
        dx = _direction_pixels - np.tile(center_pixel, (_direction_pixels.shape[0], 1))
        distances = numpy.linalg.norm(dx, axis=1)
        nearest_pixel = _direction_pixels[np.argmin(distances), :]
        furthest_pixel = _direction_pixels[np.argmax(distances), :]
        rise = furthest_pixel[0] - nearest_pixel[0]
        run = furthest_pixel[1] - nearest_pixel[1]
        self._cardinal_direction[direction] = [rise, run]

        if validate:
            fig, ax = plt.subplots()
            p = ax.pcolormesh(self.skymap['az'])
            plt.colorbar(p, ax=ax)
            ax.scatter(_direction_pixels[:, 1], _direction_pixels[:, 0])
            ax.set_title(f'{direction} | {rise=}, {run=}')
            plt.show()

        return rise, run

    def _create_animation(
        self,
        image_paths: List[pathlib.Path],
        movie_save_path: pathlib.Path,
        ffmpeg_params: dict,
        overwrite: bool,
    ):
        """
        Helper function to stich together images into an animation using ffmpeg.

        Parameters
        ----------
        image_save_dir: pathlib.Path
            The directory where the individual images are saved.
        ffmpeg_params: dict
            The additional/overwitten ffmpeg output parameters. The default parameters are:
            framerate=10, crf=25, vcodec=libx264, pix_fmt=yuv420p, preset=slower.
        movie_file_name: str
            The movie file name.
        overwrite: bool
            Overwrite the movie.
        """
        _ffmpeg_params = {
            'framerate': 10,
            'crf': 25,
            'vcodec': 'libx264',
            'pix_fmt': 'yuv420p',
            'preset': 'slower',
        }
        # Add or change the ffmpeg_params's key:values with ffmpeg_params
        _ffmpeg_params.update(ffmpeg_params)

        try:
            movie_obj = ffmpeg.input(
                str(image_paths[0].parent / "%06d.png"),
                pattern_type='sequence',
                # Pop so it won't be passed into movie_obj.output().
                framerate=_ffmpeg_params.pop('framerate'),
            )
            movie_obj.output(str(movie_save_path), **_ffmpeg_params).run(overwrite_output=overwrite)
            print(f'Animation saved to {movie_save_path}')
        except FileNotFoundError as err:
            if '[WinError 2] The system cannot find the file specified' in str(err):
                raise FileNotFoundError("Windows doesn't have ffmpeg installed.") from err
        return

    def __repr__(self):
        """
        A machine-readable representation of Imager.
        """
        params = (
            f'file_info={self.file_info}, skymap={self.skymap}, '
            f'meta={self.meta}, plot_settings={self.plot_settings}'
        )
        return f'{self.__class__.__qualname__}(' + params + ')'

    def _rgb_replacer(self, image):
        """
        Replace some RGB channels with 0. We can't use NaN, as plt.imshow changed the behavior in
        matplotlib >= 3.10.1 (https://matplotlib.org/stable/api/prev_api_changes/api_changes_3.10.1.html).
        """
        if not set(self.meta['colors']).issubset('rgb'):
            raise ValueError(" The only valid characters for the colors kwarg are 'r', 'g', 'b' ")
        
        if (*self.meta['colors'],) == ['r', 'g', 'b']:
            return image

        else:
            image = image.astype(float)
            # tests if color is selected, if not selected, then add nan values to array in lieu of color
            if 'r' not in (*self.meta['colors'],):
                # takes the shape of c, excluding the last index (-1) and replaces that matrix with nans
                image[..., 0] = np.full(np.shape(image)[:-1], 0)
            if 'g' not in (*self.meta['colors'],):
                image[..., 1] = np.full(np.shape(image)[:-1], 0)
            if 'b' not in (*self.meta['colors'],):
                image[..., 2] = np.full(np.shape(image)[:-1], 0)
        return image

class Skymap_Cleaner:
    def __init__(self, lon_grid, lat_grid, el_grid):
        """
        Clean the skymap by applying two steps: 1) Mask the lon_grid and lat_grid
        variables with np.nan if they are below min_elevation, and 2) replace all
        np.nan values in lon_grid, lat_grid with the closest valid value.
        """
        # Copy variables to not modify original np.arrays.
        self._lon_grid = lon_grid.copy()
        self._lat_grid = lat_grid.copy()
        self._el_grid = el_grid
        return
    
    def mask_elevation(self, min_elevation):
        """
        Mask low elevation pixels.
        """
        if self._el_grid.shape == self._lon_grid.shape:
            _el_grid = self._el_grid.copy()
        elif (self._el_grid.shape[0]+1 == self._lon_grid.shape[0]) and (self._el_grid.shape[1]+1 == self._lon_grid.shape[1]):
            # If the lat/lon grid is defined at pixel vertices while the el grid defines the pixel centers.
            # Here we will pretend that the el_grid is defined at pixel vertices by appending one row and column
            _el_grid = self._el_grid.copy()
            _el_grid = np.concatenate((_el_grid, _el_grid[-1, :].reshape(1, _el_grid.shape[1])), axis=0)
            _el_grid = np.concatenate((_el_grid, _el_grid[:, -1].reshape(_el_grid.shape[0], 1)), axis=1)
        else:
            raise ValueError(
                f'Can not apply elevation mask to lat/lon skymap with shape {self._lon_grid.shape} and'
                f'the elevation skymap shape {self._el_grid.shape}. The shapes must be equal or the '
                f'elevation skymap must be one less than the lon_grid.'
                )
        idh = np.where(np.isnan(_el_grid) | (_el_grid < min_elevation))
        self._lon_grid[idh] = np.nan
        self._lat_grid[idh] = np.nan
        _el_grid[idh] = np.nan
        return self._lon_grid, self._lat_grid, _el_grid

    def remove_nans(self):
        """
        Remove any NaN values from the lat and lon skymaps. We need to do this
        since pcolormesh can't handle any nan values in the x or y arrays.
         
        This method segments the skymaps in angular coordinates, (e.g. (0, 10), 
        (10, 20), ...), finds the points in x & y that are between those angles, 
        and reassigns those invalid values to the valid point with the lowest 
        elevation in that segment.
        """
        if np.all(np.isfinite(self._lon_grid)) and np.all(np.isfinite(self._lat_grid)):
            return self._lon_grid, self._lat_grid
            
        # The angular slices rotate around this point. 
        center_index = np.unravel_index(np.nanargmax(self._el_grid), self._el_grid.shape)
        num=100

        angles = np.linspace(0, 2*np.pi, num=num)
        # We need to set up 2 sets of grids since U Calgary's (lat, lon) skymap
        # dimentions are 1 greater than the (elevation, azimuth) skymaps.
        xx_geodetic, yy_geodetic = np.meshgrid(
            np.arange(self._lon_grid.shape[0]), 
            np.arange(self._lon_grid.shape[1]), 
            indexing='ij'  # So that the shapes of x, y, xx, and yy are the same.
            )
        xx_elevation, yy_elevation = np.meshgrid(
            np.arange(self._el_grid.shape[0]), 
            np.arange(self._el_grid.shape[1]), 
            indexing='ij'  # So that the shapes of x, y, xx, and yy are the same.
            )
        
        for (start_angle, end_angle) in zip(angles[:-1], angles[1:]):
            start_slope = np.tan(start_angle)
            start_y_int = center_index[1] - start_slope*center_index[0]

            end_slope = np.tan(end_angle)
            end_y_int = center_index[1] - end_slope*center_index[0]

            if (start_angle < np.pi/2) or (start_angle > 3*np.pi/2):
                start_op = operator.le
            else:
                start_op = operator.ge
            
            if (end_angle < np.pi/2) or (end_angle > 3*np.pi/2):
                end_op = operator.ge
            else:
                end_op = operator.le

            geodetic_slice_invalid_indices = np.where(
                    start_op(yy_geodetic, start_slope*xx_geodetic + start_y_int) &
                    end_op(yy_geodetic, end_slope*xx_geodetic + end_y_int) &
                    (np.isnan(self._lon_grid) | np.isnan(self._lat_grid))
                )

            if self._lat_grid.shape == self._el_grid.shape:
                elevation_slice_valid_indices = np.where(
                    start_op(yy_elevation, start_slope*xx_elevation + start_y_int) &
                    end_op(yy_elevation, end_slope*xx_elevation + end_y_int) &
                    np.isfinite(self._el_grid) & 
                    np.isfinite(self._lon_grid) &
                    np.isfinite(self._lat_grid)
                )
            else:
                elevation_slice_valid_indices = np.where(
                    start_op(yy_elevation, start_slope*xx_elevation + start_y_int) &
                    end_op(yy_elevation, end_slope*xx_elevation + end_y_int) &
                    np.isfinite(self._el_grid) & 
                    np.isfinite(self._lon_grid[:-1, :-1]) &
                    np.isfinite(self._lat_grid[:-1, :-1])
                )
            min_el_slice_index = np.argmin(self._el_grid[elevation_slice_valid_indices])
            self._lon_grid[geodetic_slice_invalid_indices] = self._lon_grid[
                elevation_slice_valid_indices[0][min_el_slice_index],
                elevation_slice_valid_indices[1][min_el_slice_index]
                ]
            self._lat_grid[geodetic_slice_invalid_indices] = self._lat_grid[
                elevation_slice_valid_indices[0][min_el_slice_index],
                elevation_slice_valid_indices[1][min_el_slice_index]
                ]
               
        if np.any(~np.isfinite(self._lon_grid)) or np.any(~np.isfinite(self._lat_grid)):
            raise ValueError(
                'Either the lat or lon skymap still contain nan values '
                'which are not allowed. Please submit a GitHub bug report if you '
                'encounter this in the wild'
                )
        return self._lon_grid, self._lat_grid


def _haversine(
    lat1: np.array, lon1: np.array, lat2: np.array, lon2: np.array, r: float = 1
) -> np.array:
    """
    _haversine distance equation.

    Parameters
    ----------
    lat1, lat2: np.array
        The latitude of points 1 and 2 in units of degrees. Can be n-dimensional.
    lon1, lon2: np.array
        The longitude of points 1 and 2 in units of degrees. Can be n-dimensional.
    r: float
        The sphere radius.
    """
    assert (
        lat1.shape == lon1.shape == lat2.shape == lon2.shape
    ), 'All input arrays must have the same shape.'
    lat1_rad = np.deg2rad(lat1)
    lat2_rad = np.deg2rad(lat2)
    lon1_rad = np.deg2rad(lon1)
    lon2_rad = np.deg2rad(lon2)

    d = (
        2
        * r
        * np.arcsin(
            np.sqrt(
                np.sin((lat1_rad - lat2_rad) / 2) ** 2
                + np.cos(lat1_rad) * np.cos(lat2_rad) * np.sin((lon2_rad - lon1_rad) / 2) ** 2
            )
        )
    )
    return d


"""
As part of the Canadian Space Agency's Geospace Observatory (GO) Canada initiative, the Auroral Imaging Group operates an array of redline auroral all-sky imagers designed by the AIG for the RISR-C CFI project. These systems are specifically designed to capture image data from only the 630.0nm wavelength (red coloured aurora). The Redline Geospace Observatory (REGO) array consists of 9 systems and are co-located with other auroral instrumentation that we operate to maximize the scientific impact. These imaging systems were deployed in late 2014 and continue to operate.
"""

from datetime import datetime, timedelta
import functools
import re
import warnings
import pathlib
import copy
import os
import dateutil.parser
import gzip
import signal
from multiprocessing import Pool

import numpy as np
import pandas as pd
import scipy.io
import matplotlib.colors

import asilib
import asilib.asi.themis as themis
import asilib.utils as utils
import asilib.download as download
import asilib.skymap


pgm_base_url = 'https://data.phys.ucalgary.ca/sort_by_project/GO-Canada/REGO/stream0/'
skymap_base_url = 'https://data.phys.ucalgary.ca/sort_by_project/GO-Canada/REGO/skymap/'
local_base_dir = asilib.config['ASI_DATA_DIR'] / 'rego'


def rego(
    location_code: str,
    time: utils._time_type = None,
    time_range: utils._time_range_type = None,
    alt: int = 110,
    custom_alt: bool = False,
    redownload: bool = False,
    missing_ok: bool = True,
    load_images: bool = True,
    acknowledge: bool = True,
    dark: bool = False,
    imager=asilib.Imager,
) -> asilib.Imager:
    """
    Create an Imager instance with the REGO ASI images and skymaps.

    Redline Emission Geospace Observatory (REGO) data is courtesy of Space Environment Canada (space-environment.ca). Use of the data must adhere to the rules of the road for that dataset. Please see below for the required data acknowledgement. Any questions about the REGO instrumentation or data should be directed to the University of Calgary, Emma Spanswick (elspansw@ucalgary.ca) and/or Eric Donovan (edonovan@ucalgary.ca).

    “The Redline Emission Geospace Observatory (REGO) is a joint Canada Foundation for Innovation and Canadian Space Agency project developed by the University of Calgary. REGO is operated and maintained by Space Environment Canada with the support of the Canadian Space Agency (CSA) [23SUGOSEC].”

    Parameters
    ----------
    location_code: str
        The ASI's location code (four letters).
    time: str or datetime.datetime
        A time to look for the ASI data at. Either time or time_range
        must be specified (not both or neither).
    time_range: list of str or datetime.datetime
        A length 2 list of string-formatted times or datetimes to bracket
        the ASI data time interval.
    alt: int
        The reference skymap altitude, in kilometers.
    custom_alt: bool
        If True, asilib will calculate (lat, lon) skymaps assuming a spherical Earth. Otherwise, it will use the official skymaps (Courtesy of University of Calgary).

        .. note::
        
            The spherical model of Earth's surface is less accurate than the oblate spheroid geometrical representation. Therefore, there will be a small difference between these and the official skymaps.
    redownload: bool
        If True, will download the data from the internet, regardless of
        wether or not the data exists locally (useful if the data becomes
        corrupted).
    missing_ok: bool
        Wether to allow missing data files inside time_range (after searching
        for them locally and online).
    load_images: bool
        Create an Imager object without images. This is useful if you need to
        calculate conjunctions and don't need to download or load unnecessary data.
    acknowledge: bool
        If True, prints the acknowledgment statement for REGO. 
    imager: asilib.Imager
        Controls what Imager instance to return, asilib.Imager by default. This
        parameter is useful if you need to subclass asilib.Imager.

    Returns
    -------
    :py:meth:`~asilib.imager.Imager`
        The THEMIS Imager instance.

    Examples
    --------
    >>> # Plot a single image.
    >>> from datetime import datetime
    >>> import matplotlib.pyplot as plt
    >>> import asilib.asi
    >>> import asilib.map
    >>> location_code = 'RANK'
    >>> time = datetime(2017, 9, 15, 2, 34, 0)
    >>> alt_km = 110
    >>> fig = plt.figure(figsize=(10, 6))
    >>> ax = fig.add_subplot(121)
    >>> bx = asilib.map.create_map(fig_ax=(fig, 122), lon_bounds = (-102, -82), lat_bounds = (58, 68))
    >>> asi = asilib.asi.rego(location_code, time=time, alt=alt_km)
    >>> asi.plot_fisheye(ax=ax)
    >>> asi.plot_map(ax=bx)
    >>> plt.tight_layout()
    >>> plt.show()

    Returns
    -------
    :py:meth:`~asilib.imager.Imager`
        A REGO ASI instance with the time stamps, images, skymaps, and metadata.
    """
    if time is not None:
        time = utils.validate_time(time)
    else:
        time_range = utils.validate_time_range(time_range)

    local_pgm_dir = local_base_dir / 'images' / location_code.lower()
    
    file_search_str = functools.partial(_glob_filename, dark=dark)

    if load_images:
        # Download and find image data
        file_paths = themis._get_pgm_files(
            'rego',
            location_code,
            time,
            time_range,
            pgm_base_url,
            local_pgm_dir,
            redownload,
            missing_ok,
            file_search_str=file_search_str
        )

        start_times = len(file_paths) * [None]
        end_times = len(file_paths) * [None]
        for i, file_path in enumerate(file_paths):
            date_match = re.search(r'\d{8}_\d{4}', file_path.name)
            start_times[i] = datetime.strptime(date_match.group(), '%Y%m%d_%H%M')
            end_times[i] = start_times[i] + timedelta(minutes=1)
        file_info = {
            'path': file_paths,
            'start_time': start_times,
            'end_time': end_times,
            'loader': _load_rego_pgm,
        }
    else:
        file_info = {
            'path': [],
            'start_time': [],
            'end_time': [],
            'loader': [],
        }
    if time_range is not None:
        file_info['time_range'] = time_range
    else:
        file_info['time'] = time

    # Download and find the appropriate skymap
    if time is not None:
        _time = time
    else:
        _time = time_range[0]
    _skymap = rego_skymap(location_code, _time, redownload=redownload)
    
    if custom_alt==False:
        alt_index = np.where(_skymap['FULL_MAP_ALTITUDE'] / 1000 == alt)[0] #Compares the altitudes versus the ones provided by default and chooses the correct index that correlates to the chosen alitudes
        assert (
            len(alt_index) == 1
        ), f'{alt} km is not in the valid skymap altitudes: {_skymap["FULL_MAP_ALTITUDE"]/1000} km. If you want a custom altitude with less percision, please use the custom_alt keyword'
        alt_index = alt_index[0]
        lat=_skymap['FULL_MAP_LATITUDE'][alt_index, :, :] #selects lat lon coordinates from data provided in skymap
        lon=_skymap['FULL_MAP_LONGITUDE'][alt_index, :, :]
    else:
        lat,lon = asilib.skymap.geodetic_skymap( #Spherical projection for lat lon coordinates
            (float(_skymap['SITE_MAP_LATITUDE']), float(_skymap['SITE_MAP_LONGITUDE']), float(_skymap['SITE_MAP_ALTITUDE']) / 1e3),
            _skymap['FULL_AZIMUTH'],
            _skymap['FULL_ELEVATION'],
            alt
            )

    skymap = {
        'lat': lat,
        'lon': lon,
        'alt': alt,
        'el': _skymap['FULL_ELEVATION'],
        'az': _skymap['FULL_AZIMUTH'],
        'path': _skymap['PATH'],
    }

    meta = {
        'array': 'REGO',
        'location': location_code.upper(),
        'lat': float(_skymap['SITE_MAP_LATITUDE']),
        'lon': float(_skymap['SITE_MAP_LONGITUDE']),
        'alt': float(_skymap['SITE_MAP_ALTITUDE']) / 1e3,
        'cadence': 3,
        'resolution': (512, 512),
        'acknowledgment':(
            'Redline Emission Geospace Observatory (REGO) data is courtesy of Space Environment '
            'Canada (space-environment.ca). Use of the data must adhere to the rules of the road '
            'for that dataset. Please see below for the required data acknowledgement. Any '
            'questions about the REGO instrumentation or data should be directed to the University '
            'of Calgary, Emma Spanswick (elspansw@ucalgary.ca) and/or Eric Donovan '
            '(edonovan@ucalgary.ca).\n\n"The Redline Emission Geospace Observatory (REGO) is a joint '
            'Canada Foundation for Innovation and Canadian Space Agency project developed by the '
            'University of Calgary. REGO is operated and maintained by Space Environment Canada '
            'with the support of the Canadian Space Agency (CSA) [23SUGOSEC]."'
            )
    }
    plot_settings = {
        'color_map': matplotlib.colors.LinearSegmentedColormap.from_list('black_to_red', ['k', 'r']),
        'color_bounds':(300, 1000)
    }

    if acknowledge and ('rego' not in asilib.config['ACKNOWLEDGED_ASIS']):
        print(meta['acknowledgment'])
        asilib.config['ACKNOWLEDGED_ASIS'].append('rego')   
    return imager(file_info, meta, skymap, plot_settings=plot_settings)


def rego_info() -> pd.DataFrame:
    """
    Returns a pd.DataFrame with the REGO ASI names and locations.

    Returns
    -------
    pd.DataFrame
        A table of REGO imager names and locations.
    """
    path = pathlib.Path(asilib.__file__).parent / 'data' / 'asi_locations.csv'
    df = pd.read_csv(path)
    df = df[df['array'] == 'REGO']
    return df.reset_index(drop=True)


def rego_skymap(location_code: str, time: utils._time_type, redownload: bool = False) -> dict:
    """
    Load a REGO ASI skymap file.

    Parameters
    ----------
    location_code: str
        The four character location name.
    time: str or datetime.datetime
        A ISO-formatted time string or datetime object. Must be in UT time.
    redownload: bool
        Redownload the file.

    Returns
    -------
    dict
        The skymap.
    """
    time = utils.validate_time(time)
    local_dir = local_base_dir / 'skymaps' / location_code.lower()
    local_dir.mkdir(parents=True, exist_ok=True)
    skymap_top_url = skymap_base_url + location_code.lower() + '/'

    if redownload:
        # Delete any existing skymap files.
        local_skymap_paths = pathlib.Path(local_dir).rglob(f'*skymap_{location_code.lower()}*.sav')
        for local_skymap_path in local_skymap_paths:
            os.unlink(local_skymap_path)
        local_skymap_paths = _download_all_skymaps(
            location_code, skymap_top_url, local_dir, redownload=redownload
        )

    else:
        local_skymap_paths = sorted(
            pathlib.Path(local_dir).rglob(f'*skymap_{location_code.lower()}*.sav')
        )
        # TODO: Add a check to periodically redownload the skymap data, maybe once a month?
        if len(local_skymap_paths) == 0:
            local_skymap_paths = _download_all_skymaps(
                location_code, skymap_top_url, local_dir, redownload=redownload
            )

    skymap_filenames = [local_skymap_path.name for local_skymap_path in local_skymap_paths]
    skymap_file_dates = []
    for skymap_filename in skymap_filenames:
        date_match = re.search(r'\d{8}', skymap_filename)
        skymap_file_dates.append(datetime.strptime(date_match.group(), '%Y%m%d'))

    # Find the skymap_date that is closest and before time.
    # For reference: dt > 0 when time is after skymap_date.
    dt = np.array([(time - skymap_date).total_seconds() for skymap_date in skymap_file_dates])
    dt[dt < 0] = np.inf  # Mask out all skymap_dates after time.
    if np.all(~np.isfinite(dt)):
        # Edge case when time is before the first skymap_date.
        closest_index = 0
        warnings.warn(
            f'The requested skymap time={time} for REGO-{location_code.upper()} is before first '
            f'skymap file dated: {skymap_file_dates[0]}. This skymap file will be used.'
        )
    else:
        closest_index = np.nanargmin(dt)
    skymap_path = local_skymap_paths[closest_index]
    skymap = _load_skymap(skymap_path)
    return skymap


def _download_all_skymaps(location_code, url, save_dir, redownload):
    d = download.Downloader(url, headers={'User-Agent':'asilib'})
    # Find the dated subdirectories
    ds = d.ls(f'{location_code.lower()}')

    save_paths = []
    for d_i in ds:
        ds = d_i.ls(f'*skymap_{location_code.lower()}*.sav')
        for ds_j in ds:
            save_paths.append(ds_j.download(save_dir, redownload=redownload))
    return save_paths


def _load_skymap(skymap_path):
    """
    A helper function to load a REGO skymap and transform it.
    """
    # Load the skymap file and convert it to a dictionary.
    skymap_file = scipy.io.readsav(str(skymap_path), python_dict=True)['skymap']
    skymap_dict = {key: copy.copy(skymap_file[key][0]) for key in skymap_file.dtype.names}

    skymap_dict = _tranform_longitude_to_180(skymap_dict)
    skymap_dict = _flip_skymap(skymap_dict)
    skymap_dict['PATH'] = skymap_path
    return skymap_dict


def _flip_skymap(skymap):
    """
    IDL is a column-major language while Python is row-major. This function
    tranposes the 2- and 3-D arrays to make them compatible with the images
    that are saved in row-major.
    """
    for key in skymap:
        if hasattr(skymap[key], 'shape'):
            shape = skymap[key].shape
            if (len(shape) == 2) and (shape[0] == shape[1]):
                skymap[key] = skymap[key][::-1, :]  # For Az/El maps.
            elif (len(shape) == 3) and (shape[1] == shape[2]):
                skymap[key] = skymap[key][:, ::-1, :]  # For lat/lon maps
    return skymap


def _tranform_longitude_to_180(skymap):
    """
    Transform the SITE_MAP_LONGITUDE and FULL_MAP_LONGITUDE arrays from
    (0 -> 360) to (-180 -> 180).
    """
    skymap['SITE_MAP_LONGITUDE'] = np.mod(skymap['SITE_MAP_LONGITUDE'] + 180, 360) - 180

    # Don't take the modulus of NaNs
    valid_val_idx = np.where(~np.isnan(skymap['FULL_MAP_LONGITUDE']))
    skymap['FULL_MAP_LONGITUDE'][valid_val_idx] = (
        np.mod(skymap['FULL_MAP_LONGITUDE'][valid_val_idx] + 180, 360) - 180
    )
    return skymap


def _load_rego_pgm(path):
    images, meta, problematic_file_list = read(str(path))
    if len(problematic_file_list):
        raise ValueError(f'A problematic PGM file: {problematic_file_list[0]}')
    images = np.moveaxis(images, 2, 0)
    images = images[:, ::-1, :]  # Flip so north is up and .
    times = np.array(
        [
            dateutil.parser.parse(dict_i['Image request start']).replace(tzinfo=None)
            for dict_i in meta
        ]
    )
    return times, images

def _glob_filename(time, location_code, _, dark=False):
    """
    Return a file search string to pass into asilib.Downloader.
    """
    if dark:
        return f'{time.strftime("%Y%m%d_%H%M")}_{location_code.lower()}_rego*6300_dark.pgm.gz'
    else:
        return f'{time.strftime("%Y%m%d_%H%M")}_{location_code.lower()}_rego*6300.pgm.gz'


"""
TThe data reader below was developed by the University of Calgary under the 
MIT License. You can find the original source files at:  
https://github.com/ucalgary-srs/rego-imager-readfile

MIT License

Copyright (c) 2020 University of Calgary

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# globals
REGO_EXPECTED_HEIGHT = 512
REGO_EXPECTED_WIDTH = 512
REGO_DT = np.dtype("uint16")
REGO_DT = REGO_DT.newbyteorder('>')  # force big endian byte ordering


def read(file_list, workers=1, first_frame=False, no_metadata=False, quiet=False):
    """
    Read in a single PGM file or set of PGM files

    :param file_list: filename or list of filenames
    :type file_list: str
    :param workers: number of worker processes to spawn, defaults to 1
    :type workers: int, optional
    :param first_frame: only read the first frame for each file, defaults to False
    :type first_frame: bool, optional
    :param no_metadata: exclude reading of metadata (performance optimization if
                        the metadata is not needed), defaults to False
    :type no_metadata: bool, optional
    :param quiet: reduce output while reading data
    :type quiet: bool, optional

    :return: images, metadata dictionaries, and problematic files
    :rtype: numpy.ndarray, list[dict], list[dict]
    """
    # if input is just a single file name in a string, convert to a list to be fed to the workers
    if isinstance(file_list, str):
        file_list = [file_list]

    # check workers
    if (workers > 1):
        try:
            # set up process pool (ignore SIGINT before spawning pool so child processes inherit SIGINT handler)
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool = Pool(processes=workers)
            signal.signal(signal.SIGINT, original_sigint_handler)  # restore SIGINT handler
        except ValueError:
            # likely the read call is being used within a context that doesn't support the usage
            # of signals in this way, proceed without it
            pool = Pool(processes=workers)

        # call readfile function, run each iteration with a single input file from file_list
        # NOTE: structure of data - data[file][metadata dictionary lists = 1, images = 0][frame]
        data = []
        try:
            data = pool.map(functools.partial(
                __rego_readfile_worker,
                first_frame=first_frame,
                no_metadata=no_metadata,
                quiet=quiet,
            ), file_list)
        except KeyboardInterrupt:
            pool.terminate()  # gracefully kill children
            return np.empty((0, 0, 0), dtype=REGO_DT), [], []
        else:
            pool.close()
            pool.join()
    else:
        # don't bother using multiprocessing with one worker, just call the worker function directly
        data = []
        for f in file_list:
            data.append(__rego_readfile_worker(
                f,
                first_frame=first_frame,
                no_metadata=no_metadata,
                quiet=quiet,
            ))

    # derive number of frames to prepare for
    total_num_frames = 0
    image_height = REGO_EXPECTED_HEIGHT
    image_width = REGO_EXPECTED_WIDTH
    for i in range(0, len(data)):
        if (data[i][2] is True):
            continue
        total_num_frames += data[i][0].shape[2]
        image_height = data[i][0].shape[0]
        image_width = data[i][0].shape[1]

    # pre-allocate array sizes
    images = np.empty([image_height, image_width, total_num_frames], dtype=REGO_DT)
    metadata_dict_list = [{}] * total_num_frames
    problematic_file_list = []

    # populate data
    list_position = 0
    for i in range(0, len(data)):
        # check if file was problematic
        if (data[i][2] is True):
            problematic_file_list.append({
                "filename": data[i][3],
                "error_message": data[i][4],
            })
            continue

        # check if any data was read in
        if (len(data[i][1]) == 0):
            continue

        # find actual number of frames, this may differ from predicted due to dropped frames, end
        # or start of imaging
        real_num_frames = data[i][0].shape[2]

        # metadata dictionary list at data[][1]
        metadata_dict_list[list_position:list_position + real_num_frames] = data[i][1]
        images[:, :, list_position:list_position + real_num_frames] = data[i][0]  # image arrays at data[][0]
        list_position = list_position + real_num_frames  # advance list position

    # trim unused elements from predicted array sizes
    metadata_dict_list = metadata_dict_list[0:list_position]
    images = np.delete(images, range(list_position, total_num_frames), axis=2)

    # ensure entire array views as uint16
    images = images.astype(np.uint16)

    # return
    data = None
    return images, metadata_dict_list, problematic_file_list


def __rego_readfile_worker(file, first_frame=False, no_metadata=False, quiet=False):
    # init
    images = np.array([])
    metadata_dict_list = []
    is_first = True
    metadata_dict = {}
    site_uid = ""
    device_uid = ""
    problematic = False
    error_message = ""

    # check file extension to see if it's gzipped or not
    try:
        if file.endswith("pgm.gz"):
            unzipped = gzip.open(file, mode='rb')
        elif file.endswith("pgm"):
            unzipped = open(file, mode='rb')
        else:
            if (quiet is False):
                print("Unrecognized file type: %s" % (file))
            problematic = True
            error_message = "Unrecognized file type"
            try:
                unzipped.close()
            except Exception:
                pass
            return images, metadata_dict_list, problematic, file, error_message
    except Exception as e:
        if (quiet is False):
            print("Failed to open file '%s' " % (file))
        problematic = True
        error_message = "failed to open file: %s" % (str(e))
        try:
            unzipped.close()
        except Exception:
            pass
        return images, metadata_dict_list, problematic, file, error_message

    # read the file
    prev_line = None
    line = None
    while True:
        # break out depending on first_frame param
        if (first_frame is True and is_first is False):
            break

        # read a line
        try:
            prev_line = line
            line = unzipped.readline()
        except Exception as e:
            if (quiet is False):
                print("Error reading before image data in file '%s'" % (file))
            problematic = True
            metadata_dict_list = []
            images = np.array([])
            error_message = "error reading before image data: %s" % (str(e))
            try:
                unzipped.close()
            except Exception:
                pass
            return images, metadata_dict_list, problematic, file, error_message

        # break loop at end of file
        if (line == b''):
            break

        # magic number; this is not a metadata or image line, exclude
        if (line.startswith(b'P5\n')):
            continue

        # process line
        if (line.startswith(b'#"')):
            if (no_metadata is True):
                metadata_dict = {}
                metadata_dict_list.append(metadata_dict)
            else:
                # metadata lines start with #"<key>"
                try:
                    line_decoded = line.decode("ascii")
                except Exception as e:
                    # skip metadata line if it can't be decoded, likely corrupt file
                    if (quiet is False):
                        print("Error decoding metadata line: %s (line='%s', file='%s')" % (str(e), line, file))
                    problematic = True
                    error_message = "error decoding metadata line: %s" % (str(e))
                    continue

                # split the key and value out of the metadata line
                line_decoded_split = line_decoded.split('"')
                if (len(line_decoded_split) != 3):
                    if (quiet is False):
                        print("Warning: issue splitting metadata line (line='%s', file='%s')" % (line_decoded, file))
                    continue
                key = line_decoded_split[1]
                value = line_decoded_split[2].strip()

                # add entry to dictionary
                metadata_dict[key] = value

                # set the site/device uids, or inject the site and device UIDs if they are missing
                if ("Site unique ID" not in metadata_dict):
                    metadata_dict["Site unique ID"] = site_uid
                else:
                    site_uid = metadata_dict["Site unique ID"]
                if ("Imager unique ID" not in metadata_dict):
                    metadata_dict["Imager unique ID"] = device_uid
                else:
                    device_uid = metadata_dict["Imager unique ID"]

                # split dictionaries up per frame, exposure plus initial readout is
                # always the end of metadata for frame
                if (key.startswith("Exposure plus readout")):
                    metadata_dict_list.append(metadata_dict)
                    metadata_dict = {}
        elif line == b'65535\n':
            # there are 2 lines between "exposure plus read out" and the image
            # data, the first is the image dimensions and the second is the max
            # value
            #
            # check the previous line to get the dimensions of the image
            prev_line_split = prev_line.decode("ascii").strip().split()
            image_width = int(prev_line_split[0])
            image_height = int(prev_line_split[1])
            bytes_to_read = image_width * image_height * 2  # 16-bit image depth

            # read image
            try:
                # read the image size in bytes from the file
                image_bytes = unzipped.read(bytes_to_read)

                # format bytes into numpy array of unsigned shorts (2byte numbers, 0-65536),
                # effectively an array of pixel values
                image_np = np.frombuffer(image_bytes, dtype=REGO_DT)

                # change 1d numpy array into matrix with correctly located pixels
                image_matrix = np.reshape(image_np, (image_height, image_width, 1))
            except Exception as e:
                if (quiet is False):
                    print("Failed reading image data frame: %s" % (str(e)))
                metadata_dict_list.pop()  # remove corresponding metadata entry
                problematic = True
                error_message = "image data read failure: %s" % (str(e))
                continue  # skip to next frame

            # initialize image stack
            if (is_first is True):
                images = image_matrix
                is_first = False
            else:
                images = np.dstack([images, image_matrix])  # depth stack images (on 3rd axis)

    # close gzip file
    unzipped.close()

    # check to see if the image is empty
    if (images.size == 0):
        if (quiet is False):
            print("Error reading image file: found no image data")
        problematic = True
        error_message = "no image data"

    # return
    return images, metadata_dict_list, problematic, file, error_message


from __future__ import annotations  # to support the -> List[Downloader] return type
from typing import List
import pathlib
import urllib
import re
import time

import numpy as np
from bs4 import BeautifulSoup
import requests


class Downloader:
    """
    Traverses and lists the directory structure on a server and download files.

    Parameters
    ----------
    url: str
        The dataset URL. If it points to a folder it must end with a "/".
    download_dir: str or pathlib.Path
        The download directory. Must either specify here, or when you call
        Downloader.download().

    Example
    -------
    TODO: Add an example.
    """

    def __init__(self, url: str, download_dir=None, headers={}) -> None:
        self.url = url
        self.download_dir = download_dir
        self.headers = headers
        return

    def ls(self, match: str = '*') -> List[Downloader]:
        """
        List files and folders in self.url.

        Parameters
        ----------
        match: str
            An optional string pattern to match.

        Return
        ------
        list
            A list of full URLs.
        """
        matched_hrefs = self._search_hrefs(self.url, match=match)
        cls = type(self)
        downloaders = [None] * len(matched_hrefs)
        for i, matched_href in enumerate(matched_hrefs):
            new_url = urllib.parse.urljoin(self.url, matched_href, allow_fragments=True)
            downloaders[i] = cls(new_url, download_dir=self.download_dir, headers=self.headers)
        return downloaders

    def download(
        self, download_dir=None, redownload: bool = False, stream: bool = False,
        timeout: int=10, retries:int=3
    ) -> pathlib.Path:
        """
        Downloads file from self.url to the download_dir directory.

        Parameters
        ----------
        download_dir: str or pathlib.Path
            The parent directory where to save the data to. Set it either
            here or when initializing the class. This
        redownload: bool
            Will redownload an existing file.
        stream: bool
            Download the data in one chunk if False, or break it up and download it
            in 5 Mb chunks if True.
        timeout: int
            The timeout in seconds for the requests.
        retries: int
            The number of times to retry the download if it fails due to a network error.

        Returns
        -------
        pathlib.Path
            The full path to the file.
        """
        if download_dir is None and self.download_dir is None:
            raise ValueError(
                f'download_dir kwarg needs to be set either '
                f'in Downloader() or Downloader.download.'
            )
        if download_dir is not None:
            self.download_dir = download_dir

        self.download_dir = pathlib.Path(self.download_dir)
        self.download_dir.mkdir(parents=True, exist_ok=True)
        file_name = pathlib.Path(self.url).name
        download_path = self.download_dir / file_name

        if (download_path.exists()) and (not redownload):
            return download_path

        if stream:
            with requests.Session() as s:
                r = s.get(self.url, stream=True, timeout=timeout, headers=self.headers)
            r.raise_for_status()
            file_size = int(r.headers.get('content-length'))
            downloaded_bites = 0

            megabyte = 1024 * 1024
            try:
                with open(download_path, 'wb') as f:
                    for data in r.iter_content(chunk_size=10 * megabyte):
                        f.write(data)
                        # Update the downloaded % in the terminal.
                        downloaded_bites += len(data)
                        download_percent = round(100 * downloaded_bites / file_size)
                        download_str = "#" * (download_percent // 5)
                        print(
                            f'Downloading {file_name}: |{download_str:<20}| {download_percent}%',
                            end='\r',
                        )
                print()  # Add a newline
            except (Exception, KeyboardInterrupt, SystemExit) as err:
                download_path.unlink()
                raise RuntimeError(
                    f'Download interrupted. Partially downloaded file ' f'{download_path} deleted.'
                ) from err
        else:
            response = self.retry_download(self.url, timeout=timeout, retries=retries)
            with open(download_path, 'wb') as f:
                f.write(response.content)
            print(f'Downloaded {file_name}.')
        return download_path

    @property
    def name(self):
        """
        Get the url filename
        """
        _url = pathlib.Path(self.url)
        if _url.suffix != '':
            # url points to a filename. pathlib is not designed for
            # urls, but this is the easiest way to get the name.
            return _url.name
        else:
            return None

    def _search_hrefs(self, url: str, match: str = '*') -> List[str]:
        """
        Given a url string, find all hyper references matching the
        string match. The re module does the matching.

        Parameters
        ----------
        url: str
            A url
        match: str (optional)
            The regex match to compare the hyper references to. The default is
            to match everything (a wildcard)

        Returns
        -------
        List(str)
            A list of hyper references that matched the match string.

        Raises
        ------
        FileNotFoundError
            If no hyper references were found.
        """
        with requests.Session() as s:
            request = self.retry_download(url, timeout=5, retries=3)
        request.raise_for_status()
        soup = BeautifulSoup(request.content, 'html.parser')

        match = match.replace('*', '.*')  # regex wildcard
        href_objs = soup.find_all('a', href=re.compile(match))
        # I found "?" to be in the column names so we should ignore them.
        matched_hrefs = [
            href_obj['href'] for href_obj in href_objs if href_obj['href'].find('?') == -1
        ]
        if len(matched_hrefs) == 0:
            raise FileNotFoundError(
                f'The url {url} does not contain any hyper '
                f'references containing the match kwarg="{match}".'
            )
        return matched_hrefs
    
    def retry_download(self, url: str, timeout: int=10, retries:int=3):
        """
        Retry downloading from a url.
        """
        n_tries = 0
        sleep_times = np.logspace(0, 1, retries)
        
        while n_tries < retries:
            try:
                response = requests.get(url, timeout=timeout, headers=self.headers)
                break
            except (requests.ConnectionError, requests.Timeout) as err:
                n_tries += 1
                print(f'Download failed for {url}, attempt {n_tries}/{retries}.')
                if n_tries == retries:
                    raise
                time.sleep(sleep_times[n_tries])
        return response

    def __repr__(self) -> str:
        params = f'{self.url}, download_dir={self.download_dir},'
        return f'{self.__class__.__qualname__}(' + params + ')'

    def __str__(self) -> str:
        return (
            f'{self.__class__.__qualname__} with url={self.url} and '
            f'download_dir={self.download_dir}'
        )



from datetime import datetime

import asilib.asi
import asilib.map

time_range = (datetime(2015, 3, 26, 6, 7), datetime(2015, 3, 26, 6, 12))
location_code = 'FSMI'

asi = asilib.asi.themis(location_code, time_range=time_range, alt=110)

lat_bounds = (asi.meta['lat'] - 7, asi.meta['lat'] + 7)
lon_bounds = (asi.meta['lon'] - 20, asi.meta['lon'] + 20)
ax = asilib.map.create_map(lon_bounds=lon_bounds, lat_bounds=lat_bounds)

asi.animate_map(ax=ax)

print(f'Animation saved in {asilib.config["ASI_DATA_DIR"] / "animations" / asi.animation_name}')
